#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo para obtener CAE/CAEA, c�digo de autorizaci�n electr�nico
webservice WSFEv1 de AFIP (Factura Electr�nica Nacional - Proyecto
Version 1 - 2.10) Seg�n RG 2485/08, RG 2757/2010, RG 2904/2010 y
RG2926/10 (CAE anticipado), RG 3067/2011 (RS - Monotributo), RG 3571/2013
(Responsables inscriptos IVA), RG 3668/2014 (Factura A IVA F.8001),
RG 3749/2015 (R.I. y exentos) RG 4004-E Alquiler de inmuebles con
destino casa habitaci�n). RG 4109-E Venta de bienes muebles registrables.
M�s info:
http://www.sistemasagiles.com.ar/trac/wiki/ProyectoWSFEv1 """

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010-2017 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.20a"

import datetime
# import decimal
import os
import sys
from .utils import (
    verifica, inicializar_y_capturar_excepciones, BaseWS, get_install_dir
)
import logging

log = logging.getLogger()
HOMO = False                    # solo homologaci�n
TYPELIB = False                 # usar librer�a de tipos (TLB)
LANZAR_EXCEPCIONES = False      # valor por defecto: True

# WSDL = "https://www.sistemasagiles.com.ar/simulador/wsfev1/call/soap?WSDL=
# None"
WSDL = "https://wswhomo.afip.gov.ar/wsfev1/service.asmx?WSDL"
# WSDL = "file:///home/reingart/tmp/service.asmx.xml"


class WSFEv1(BaseWS):
    "Interfaz para el WebService de Factura Electr�nica Version 1 - 2.10"
    _public_methods_ = [
        'CrearFactura', 'AgregarIva', 'CAESolicitar',
        'AgregarTributo', 'AgregarCmpAsoc', 'AgregarOpcional',
        'AgregarComprador',
        'CompUltimoAutorizado', 'CompConsultar',
        'CAEASolicitar', 'CAEAConsultar', 'CAEARegInformativo',
        'CAEASinMovimientoInformar',
        'CAESolicitarX',
        'IniciarFacturasX', 'AgregarFacturaX', 'LeerFacturaX',
        'ParamGetTiposCbte',
        'ParamGetTiposConcepto',
        'ParamGetTiposDoc',
        'ParamGetTiposIva',
        'ParamGetTiposMonedas',
        'ParamGetTiposOpcional',
        'ParamGetTiposTributos',
        'ParamGetTiposPaises',
        'ParamGetCotizacion',
        'ParamGetPtosVenta',
        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
        'SetParametros', 'SetTicketAcceso', 'GetParametro',
        'EstablecerCampoFactura', 'ObtenerCampoFactura',
        'Dummy', 'Conectar', 'DebugLog', 'SetTicketAcceso'
    ]
    _public_attrs_ = [
        'Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'XmlRequest', 'XmlResponse', 'Version', 'Excepcion',
        'LanzarExcepciones', 'Resultado', 'Obs', 'Observaciones', 'Traceback',
        'InstallDir', 'CAE', 'Vencimiento', 'Eventos', 'Errores', 'ErrCode',
        'ErrMsg', 'Reprocesar', 'Reproceso', 'EmisionTipo', 'CAEA', 'CbteNro',
        'CbtDesde', 'CbtHasta', 'FechaCbte', 'ImpTotal', 'ImpNeto', 'ImptoLiq',
        'ImpIVA', 'ImpOpEx', 'ImpTrib',
    ]

    _reg_progid_ = "WSFEv1"
    _reg_clsid_ = "{CA0E604D-E3D7-493A-8880-F6CDD604185E}"

    if TYPELIB:
        _typelib_guid_ = '{B1D7283C-3EC2-463E-89B4-11F5228E2A15}'
        _typelib_version_ = 1, 18
        _com_interfaces_ = ['IWSFEv1']
        # _reg_class_spec_ = "wsfev1.WSFEv1"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologaci�n' or '')
    Reprocesar = True   # recuperar automaticamente CAE emitidos
    LanzarExcepciones = LANZAR_EXCEPCIONES
    factura = None
    facturas = None

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.CAEA = \
            self.Vencimiento = ''
        self.CbteNro = self.CbtDesde = self.CbtHasta = self.PuntoVenta = None
        self.ImpTotal = self.ImpIVA = self.ImpOpEx = self.ImpNeto = \
            self.ImptoLiq = self.ImpTrib = None
        self.EmisionTipo = self.Periodo = self.Orden = ""
        self.FechaCbte = self.FchVigDesde = self.FchVigHasta = \
            self.FchTopeInf = self.FchProceso = ""

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        if 'Errors' in ret:
            errores = ret['Errors']
            for error in errores:
                self.Errores.append("%s: %s" % (
                    error['Err']['Code'],
                    error['Err']['Msg'],
                ))
            self.ErrCode = ' '.join(
                [str(error['Err']['Code']) for error in errores]
            )
            self.ErrMsg = '\n'.join(self.Errores)
        if 'Events' in ret:
            events = ret['Events']
            self.Eventos = [
                '%s: %s' % (
                    evt['Evt']['Code'], evt['Evt']['Msg']
                ) for evt in events
            ]

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        log.warn("### Calling Metodo Dummy para verificacion de \
        funcionamiento de la infraestructura (FEDummy)")
        result = self.client.FEDummy()['FEDummyResult']
        log.warn("FEDummy Result: " + str(result))
        self.AppServerStatus = result.get('AppServer')
        self.DbServerStatus = result.get('DbServer')
        self.AuthServerStatus = result.get('AuthServer')
        return True

    # los siguientes m�todos no est�n decorados para no limpiar propiedades

    def CrearFactura(
        self, concepto=1, tipo_doc=80, nro_doc="", tipo_cbte=1, punto_vta=0,
        cbt_desde=0, cbt_hasta=0, imp_total=0.00, imp_tot_conc=0.00,
        imp_neto=0.00, imp_iva=0.00, imp_trib=0.00, imp_op_ex=0.00,
        fecha_cbte="", fecha_venc_pago=None, fecha_serv_desde=None,
        fecha_serv_hasta=None,   # --
        moneda_id="PES", moneda_ctz="1.0000", caea=None, **kwargs
    ):
        "Creo un objeto factura (interna)"
        # Creo una factura electronica de exportaci�n
        fact = {
            'tipo_doc': tipo_doc, 'nro_doc':  nro_doc,
            'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
            'cbt_desde': cbt_desde, 'cbt_hasta': cbt_hasta,
            'imp_total': imp_total, 'imp_tot_conc': imp_tot_conc,
            'imp_neto': imp_neto, 'imp_iva': imp_iva,
            'imp_trib': imp_trib, 'imp_op_ex': imp_op_ex,
            'fecha_cbte': fecha_cbte,
            'fecha_venc_pago': fecha_venc_pago,
            'moneda_id': moneda_id, 'moneda_ctz': moneda_ctz,
            'concepto': concepto,
            'cbtes_asoc': [],
            'tributos': [],
            'iva': [],
            'opcionales': [],
            'compradores': [],
        }
        if fecha_serv_desde:
            fact['fecha_serv_desde'] = fecha_serv_desde
        if fecha_serv_hasta:
            fact['fecha_serv_hasta'] = fecha_serv_hasta
        if caea:
            fact['caea'] = caea

        self.factura = fact
        return True

    def EstablecerCampoFactura(self, campo, valor):
        if campo in self.factura or campo in (
            'fecha_serv_desde', 'fecha_serv_hasta', 'caea', 'fch_venc_cae'
        ):
            self.factura[campo] = valor
            return True
        else:
            return False

    def AgregarCmpAsoc(self, tipo=1, pto_vta=0, nro=0, cuit=None, fecha=None, 
        **kwargs):
        "Agrego un comprobante asociado a una factura (interna)"
        cmp_asoc = {
            'tipo': tipo,
            'pto_vta': pto_vta,
            'nro': nro
        }
        if cuit is not None:
            cmp_asoc['cuit'] = cuit
        if fecha is not None:
            cmp_asoc['fecha'] = fecha
        self.factura['cbtes_asoc'].append(cmp_asoc)
        return True

    def AgregarTributo(
        self,
        tributo_id=0,
        desc="",
        base_imp=0.00,
        alic=0,
        importe=0.00,
        **kwarg
    ):
        "Agrego un tributo a una factura (interna)"
        tributo = {
            'tributo_id': tributo_id,
            'desc': desc,
            'base_imp': base_imp,
            'alic': alic,
            'importe': importe
        }
        self.factura['tributos'].append(tributo)
        return True

    def AgregarIva(self, iva_id=0, base_imp=0.0, importe=0.0, **kwarg):
        "Agrego un tributo a una factura (interna)"
        iva = {
            'iva_id': iva_id,
            'base_imp': base_imp,
            'importe': importe
        }
        self.factura['iva'].append(iva)
        return True

    def AgregarOpcional(self, opcional_id=0, valor="", **kwarg):
        "Agrego un dato opcional a una factura (interna)"
        op = {
            'opcional_id': opcional_id,
            'valor': valor
        }
        self.factura['opcionales'].append(op)
        return True

    def AgregarComprador(
        self, doc_tipo=80, doc_nro=0, porcentaje=100.00, **kwarg
    ):
        "Agrego un comprador a una factura (interna) RG 4109-E bienes muebles"
        comp = {
            'doc_tipo': doc_tipo,
            'doc_nro': doc_nro,
            'porcentaje': porcentaje
        }
        self.factura['compradores'].append(comp)
        return True

    def ObtenerCampoFactura(self, *campos):
        "Obtener el valor devuelto de AFIP para un campo de factura"
        # cada campo puede ser una clave string (dict) o una posici�n (list)
        ret = self.factura
        for campo in campos:
            if isinstance(ret, dict) and isinstance(campo, str):
                ret = ret.get(campo)
            elif isinstance(ret, list) and len(ret) > campo:
                ret = ret[campo]
            else:
                self.Excepcion = "El campo %s solicitado no existe" % campo
                ret = None
            if ret is None:
                break
        return str(ret)

    # metodos principales para llamar remotamente a AFIP:

    @inicializar_y_capturar_excepciones
    def CAESolicitar(self):
        """M�todo de autorizaci�n de comprobantes electr�nicos por CAE
        (FECAESolicitar)"""
        log.warn("### Calling Metodo de autorizacion de comprobantes \
                electronicos por CAE (FECAESolicitar)")
        f = self.factura
        log.warn("### Factura: " + str(f))

        try:
            ret = self.client.FECAESolicitar(
                Auth={
                    'Token': self.Token.decode('utf8'),
                    'Sign': self.Sign.decode('utf8'),
                    'Cuit': self.Cuit
                },
                FeCAEReq={
                    'FeCabReq': {
                        'CantReg': 1,
                        'PtoVta': f['punto_vta'],
                        'CbteTipo': f['tipo_cbte']
                    },
                    'FeDetReq': [{'FECAEDetRequest': {
                        'Concepto': f['concepto'],
                        'DocTipo': f['tipo_doc'],
                        'DocNro': f['nro_doc'],
                        'CbteDesde': f['cbt_desde'],
                        'CbteHasta': f['cbt_hasta'],
                        'CbteFch': f['fecha_cbte'],
                        'ImpTotal': f['imp_total'],
                        'ImpTotConc': f['imp_tot_conc'],
                        'ImpNeto': f['imp_neto'],
                        'ImpOpEx': f['imp_op_ex'],
                        'ImpTrib': f['imp_trib'],
                        'ImpIVA': f['imp_iva'],
                        # Fechas solo se informan si Concepto in (2,3)
                        'FchServDesde': f.get('fecha_serv_desde'),
                        'FchServHasta': f.get('fecha_serv_hasta'),
                        'FchVtoPago': f['fecha_venc_pago'],
                        'MonId': f['moneda_id'],
                        'MonCotiz': f['moneda_ctz'],
                        'CbtesAsoc': f['cbtes_asoc'] and [
                            {'CbteAsoc': {
                                'Tipo': cbte_asoc['tipo'],
                                'PtoVta': cbte_asoc['pto_vta'],
                                'Nro': cbte_asoc['nro'],
                                'Cuit': cbte_asoc.get('cuit'),
                                'CbteFch': cbte_asoc.get('fecha'),
                            }}
                            for cbte_asoc in f['cbtes_asoc']] or None,
                        'Tributos': f['tributos'] and [
                            {'Tributo': {
                                'Id': tributo['tributo_id'],
                                'Desc': tributo['desc'],
                                'BaseImp': tributo['base_imp'],
                                'Alic': tributo['alic'],
                                'Importe': tributo['importe'],
                                }}
                            for tributo in f['tributos']] or None,
                        'Iva': f['iva'] and [
                            {'AlicIva': {
                                'Id': iva['iva_id'],
                                'BaseImp': iva['base_imp'],
                                'Importe': iva['importe'],
                                }}
                            for iva in f['iva']] or None,
                        'Opcionales': [
                            {'Opcional': {
                                'Id': opcional['opcional_id'],
                                'Valor': opcional['valor'],
                                }} for opcional in f['opcionales']] or None,
                        'Compradores': [
                            {'Comprador': {
                                'DocTipo': comprador['doc_tipo'],
                                'DocNro': comprador['doc_nro'],
                                'Porcentaje': comprador['porcentaje'],
                                }} for comprador in f['compradores']] or None,
                        }
                    }]
                })
        except Exception as e:
            log.warn('explotando el FECAESolicitar')
            msg = str(e)
            log.warn('msg: %s' % msg)

        result = ret['FECAESolicitarResult']
        log.warn("FECAESolicitar Result: " + str(result))
        if 'FeCabResp' in result:
            fecabresp = result['FeCabResp']
            fedetresp = result['FeDetResp'][0]['FECAEDetResponse']

            # Reprocesar en caso de error (recuperar CAE emitido anteriormente)
            if self.Reprocesar and (
                'Errors' in result or 'Observaciones' in fedetresp
            ):
                for error in result.get('Errors', []) + fedetresp.get(
                    'Observaciones', []
                ):
                    err_code = str(error.get('Err', error.get('Obs'))['Code'])
                    if fedetresp['Resultado'] == 'R' and err_code == '10016':
                        # guardo los mensajes xml originales
                        xml_request = self.client.xml_request
                        xml_response = self.client.xml_response
                        cae = self.CompConsultar(
                            f['tipo_cbte'],
                            f['punto_vta'],
                            f['cbt_desde'],
                            reproceso=True
                        )
                        if cae and self.EmisionTipo == 'CAE':
                            self.Reproceso = 'S'
                            return cae
                        self.Reproceso = 'N'
                        # reestablesco los mensajes xml originales
                        self.client.xml_request = xml_request
                        self.client.xml_response = xml_response

            self.Resultado = fecabresp['Resultado']
            # Obs:
            for obs in fedetresp.get('Observaciones', []):
                self.Observaciones.append("%(Code)s: %(Msg)s" % (obs['Obs']))
            self.Obs = '\n'.join(self.Observaciones)
            self.CAE = fedetresp['CAE'] and str(fedetresp['CAE']) or ""
            self.EmisionTipo = 'CAE'
            self.Vencimiento = fedetresp['CAEFchVto']
            # .strftime("%Y/%m/%d")
            self.FechaCbte = fedetresp.get('CbteFch', "")
            self.CbteNro = fedetresp.get('CbteHasta', 0)  # 1L
            self.PuntoVenta = fecabresp.get('PtoVta', 0)  # 4000
            self.CbtDesde = fedetresp.get('CbteDesde', 0)
            self.CbtHasta = fedetresp.get('CbteHasta', 0)
        self.__analizar_errores(result)
        return self.CAE

    @inicializar_y_capturar_excepciones
    def CompTotXRequest(self):
        "Recuperador de ultimo valor de comprobante registrado \
        (FECompUltimoAutorizado)"
        log.warn("### Calling Recuperador de cantidad máxima de registros \
        FECAESolicitar - FECAEARegInformativo (FECompTotXRequest)")
        ret = self.client.FECompTotXRequest(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            }
        )

        result = ret['FECompTotXRequestResult']
        log.warn("FECompTotXRequest Result: " + str(result))
        return result['RegXReq']

    @inicializar_y_capturar_excepciones
    def CompUltimoAutorizado(self, tipo_cbte, punto_vta):
        "Recuperador de ultimo valor de comprobante registrado \
                (FECompUltimoAutorizado)"
        log.warn("### Calling Recuperador de ultimo valor de comprobante \
        registrado (FECompUltimoAutorizado)")
        ret = self.client.FECompUltimoAutorizado(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            PtoVta=punto_vta,
            CbteTipo=tipo_cbte,
            )

        result = ret['FECompUltimoAutorizadoResult']
        log.warn("FECompUltimoAutorizado Result: " + str(result))
        self.CbteNro = result['CbteNro']
        self.__analizar_errores(result)
        return self.CbteNro is not None and str(self.CbteNro) or ''

    @inicializar_y_capturar_excepciones
    def CompConsultar(self, tipo_cbte, punto_vta, cbte_nro, reproceso=False):
        "M�todo para consultar Comprobantes Emitidos y su c�digo \
                (FECompConsultar)"
        log.warn("### Calling Metodo para consultar Comprobantes Emitidos \
        y su codigo (FECompConsultar)")

        difs = []  # si hay reproceso, verifico las diferencias con AFIP

        ret = self.client.FECompConsultar(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            FeCompConsReq={
                'CbteTipo': tipo_cbte,
                'CbteNro': cbte_nro,
                'PtoVta': punto_vta,
            })

        result = ret['FECompConsultarResult']
        log.warn("FECompConsultar Result: " + str(result))
        if 'ResultGet' in result:
            resultget = result['ResultGet']

            if reproceso:
                # verifico los campos registrados coincidan con los enviados:
                f = self.factura
                verificaciones = {
                    'Concepto': f['concepto'],
                    'DocTipo': f['tipo_doc'],
                    'DocNro': f['nro_doc'],
                    'CbteTipo': f['tipo_cbte'],
                    'CbteDesde': f['cbt_desde'],
                    'CbteHasta': f['cbt_hasta'],
                    'CbteFch': f['fecha_cbte'],
                    'ImpTotal': f['imp_total'] and (
                        float(f['imp_total']) or 0.0
                    ),
                    'ImpTotConc': f['imp_tot_conc'] and (
                        float(f['imp_tot_conc']) or 0.0
                    ),
                    'ImpNeto': f['imp_neto'] and float(f['imp_neto']) or 0.0,
                    'ImpOpEx': f['imp_op_ex'] and float(f['imp_op_ex']) or 0.0,
                    'ImpTrib': f['imp_trib'] and float(f['imp_trib']) or 0.0,
                    'ImpIVA': f['imp_iva'] and float(f['imp_iva']) or 0.0,
                    'FchServDesde': f.get('fecha_serv_desde'),
                    'FchServHasta': f.get('fecha_serv_hasta'),
                    'FchVtoPago': f.get('fecha_venc_pago'),
                    'MonId': f['moneda_id'],
                    'MonCotiz': float(f['moneda_ctz']),
                    'CbtesAsoc': [
                        {'CbteAsoc': {
                            'Tipo': cbte_asoc['tipo'],
                            'PtoVta': cbte_asoc['pto_vta'],
                            'Nro': cbte_asoc['nro'],
                            'Cuit': cbte_asoc.get('cuit'),
                        }}
                        for cbte_asoc in f['cbtes_asoc']],
                    'Tributos': [
                        {'Tributo': {
                            'Id': tributo['tributo_id'],
                            'Desc': tributo['desc'],
                            'BaseImp': float(tributo['base_imp'] or 0),
                            'Alic': float(tributo['alic'] or 0),
                            'Importe': float(tributo['importe']),
                            }}
                        for tributo in f['tributos']],
                    'Iva': [
                        {'AlicIva': {
                            'Id': iva['iva_id'],
                            'BaseImp': float(iva['base_imp']),
                            'Importe': float(iva['importe']),
                            }}
                        for iva in f['iva']],
                    'Opcionales': [
                        {'Opcional': {
                            'Id': opcional['opcional_id'],
                            'Valor': opcional['valor'],
                            }} for opcional in f['opcionales']],
                    'Compradores': [
                        {'Comprador': {
                            'DocTipo': comprador['doc_tipo'],
                            'DocNro': comprador['doc_nro'],
                            'Porcentaje': comprador['porcentaje'],
                            }} for comprador in f['compradores']],
                    }
                verifica(verificaciones, resultget.copy(), difs)
                if difs:
                    print("Diferencias:", difs)
                    self.log("Diferencias: %s" % difs)
            else:
                # guardo los datos de AFIP (reconstruyo estructura interna)
                self.factura = {
                    'concepto': resultget.get('Concepto'),
                    'tipo_doc': resultget.get('DocTipo'),
                    'nro_doc': resultget.get('DocNro'),
                    'tipo_cbte': resultget.get('CbteTipo'),
                    'punto_vta': resultget.get('PtoVta'),
                    'cbt_desde': resultget.get('CbteDesde'),
                    'cbt_hasta': resultget.get('CbteHasta'),
                    'fecha_cbte': resultget.get('CbteFch'),
                    'imp_total': resultget.get('ImpTotal'),
                    'imp_tot_conc': resultget.get('ImpTotConc'),
                    'imp_neto': resultget.get('ImpNeto'),
                    'imp_op_ex': resultget.get('ImpOpEx'),
                    'imp_trib': resultget.get('ImpTrib'),
                    'imp_iva': resultget.get('ImpIVA'),
                    'fecha_serv_desde': resultget.get('FchServDesde'),
                    'fecha_serv_hasta': resultget.get('FchServHasta'),
                    'fecha_venc_pago': resultget.get('FchVtoPago'),
                    'moneda_id': resultget.get('MonId'),
                    'moneda_ctz': resultget.get('MonCotiz'),
                    'cbtes_asoc': [
                        {
                            'tipo': cbte_asoc['CbteAsoc']['Tipo'],
                            'pto_vta': cbte_asoc['CbteAsoc']['PtoVta'],
                            'nro': cbte_asoc['CbteAsoc']['Nro'],
                            'cuit': cbte_asoc['CbteAsoc'].get('Cuit')
                        } for cbte_asoc in resultget.get('CbtesAsoc', [])
                    ],
                    'tributos': [
                        {
                            'tributo_id': tributo['Tributo']['Id'],
                            'desc': tributo['Tributo']['Desc'],
                            'base_imp': tributo['Tributo'].get('BaseImp'),
                            'alic': tributo['Tributo'].get('Alic'),
                            'importe': tributo['Tributo']['Importe'],
                        } for tributo in resultget.get('Tributos', [])
                    ],
                    'iva': [
                        {
                            'iva_id': iva['AlicIva']['Id'],
                            'base_imp': iva['AlicIva']['BaseImp'],
                            'importe': iva['AlicIva']['Importe'],
                        } for iva in resultget.get('Iva', [])
                    ],
                    'opcionales': [
                        {
                            'opcional_id': obs['Opcional']['Id'],
                            'valor': obs['Opcional']['Valor'],
                        } for obs in resultget.get('Opcionales', [])
                    ],
                    'compradores': [
                        {
                            'doc_tipo': comp['Comprador']['DocTipo'],
                            'doc_nro': comp['Comprador']['DocNro'],
                            'porcentaje': comp['Comprador']['Porcentaje'],
                        } for comp in resultget.get('Compradores', [])
                    ],
                    'cae': resultget.get('CodAutorizacion'),
                    'resultado': resultget.get('Resultado'),
                    'fch_venc_cae': resultget.get('FchVto'),
                    'obs': [
                        {
                            'code': obs['Obs']['Code'],
                            'msg': obs['Obs']['Msg'],
                        } for obs in resultget.get('Observaciones', [])
                    ],
                }

            self.FechaCbte = resultget['CbteFch']  # .strftime("%Y/%m/%d")
            self.CbteNro = resultget['CbteHasta']  # 1L
            self.PuntoVenta = resultget['PtoVta']  # 4000
            self.Vencimiento = resultget['FchVto']  # .strftime("%Y/%m/%d")
            self.ImpTotal = str(resultget['ImpTotal'])
            cod_aut = resultget['CodAutorizacion'] and \
                str(resultget['CodAutorizacion']) or ''  # 60423794871430L
            self.Resultado = resultget['Resultado']
            self.CbtDesde = resultget['CbteDesde']
            self.CbtHasta = resultget['CbteHasta']
            self.ImpTotal = resultget['ImpTotal']
            self.ImpNeto = resultget['ImpNeto']
            self.ImptoLiq = self.ImpIVA = resultget['ImpIVA']
            self.ImpOpEx = resultget['ImpOpEx']
            self.ImpTrib = resultget['ImpTrib']
            self.EmisionTipo = resultget['EmisionTipo']
            if self.EmisionTipo == 'CAE':
                self.CAE = cod_aut
            elif self.EmisionTipo == 'CAEA':
                self.CAEA = cod_aut
            # Obs:
            for obs in resultget.get('Observaciones', []):
                self.Observaciones.append("%(Code)s: %(Msg)s" % (obs['Obs']))
            self.Obs = '\n'.join(self.Observaciones)

        self.__analizar_errores(result)
        if not difs:
            return self.CAE or self.CAEA
        else:
            return ''

    @inicializar_y_capturar_excepciones
    def CAESolicitarX(self):
        "Autorizar m�ltiples facturas (CAE) en una �nica solicitud"
        # Ver CompTotXRequest -> cantidad maxima comprobantes (250)
        # verificar que hay multiples facturas:
        if not self.facturas:
            raise RuntimeError("Llamar a IniciarFacturasX y AgregarFacturaX!")
        # verificar que todas las facturas
        puntos_vta = set([f['punto_vta'] for f in self.facturas])
        tipos_cbte = set([f['tipo_cbte'] for f in self.facturas])
        if len(puntos_vta) > 1:
            raise RuntimeError("Los comprobantes deben ser del mismo pto_vta!")
        if len(tipos_cbte) > 1:
            raise RuntimeError("Los comprobantes deben tener el mismo tipo!")
        # llamar al webservice:
        ret = self.client.FECAESolicitar(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            FeCAEReq={
                'FeCabReq': {
                    'CantReg': len(self.facturas),
                    'PtoVta': puntos_vta.pop(),
                    'CbteTipo': tipos_cbte.pop()
                },
                'FeDetReq': [{'FECAEDetRequest': {
                    'Concepto': f['concepto'],
                    'DocTipo': f['tipo_doc'],
                    'DocNro': f['nro_doc'],
                    'CbteDesde': f['cbt_desde'],
                    'CbteHasta': f['cbt_hasta'],
                    'CbteFch': f['fecha_cbte'],
                    'ImpTotal': f['imp_total'],
                    'ImpTotConc': f['imp_tot_conc'],
                    'ImpNeto': f['imp_neto'],
                    'ImpOpEx': f['imp_op_ex'],
                    'ImpTrib': f['imp_trib'],
                    'ImpIVA': f['imp_iva'],
                    # Fechas solo se informan si Concepto in (2,3)
                    'FchServDesde': f.get('fecha_serv_desde'),
                    'FchServHasta': f.get('fecha_serv_hasta'),
                    'FchVtoPago': f['fecha_venc_pago'],
                    'MonId': f['moneda_id'],
                    'MonCotiz': f['moneda_ctz'],
                    'CbtesAsoc': [
                        {'CbteAsoc': {
                            'Tipo': cbte_asoc['tipo'],
                            'PtoVta': cbte_asoc['pto_vta'],
                            'Nro': cbte_asoc['nro'],
                            'Cuit': cbte_asoc.get('cuit'),
                        }}
                        for cbte_asoc in f['cbtes_asoc']] or None,
                    'Tributos': [
                        {'Tributo': {
                            'Id': tributo['tributo_id'],
                            'Desc': tributo['desc'],
                            'BaseImp': tributo['base_imp'],
                            'Alic': tributo['alic'],
                            'Importe': tributo['importe'],
                            }}
                        for tributo in f['tributos']] or None,
                    'Iva': [
                        {'AlicIva': {
                            'Id': iva['iva_id'],
                            'BaseImp': iva['base_imp'],
                            'Importe': iva['importe'],
                            }}
                        for iva in f['iva']] or None,
                    'Opcionales': [
                        {'Opcional': {
                            'Id': opcional['opcional_id'],
                            'Valor': opcional['valor'],
                            }} for opcional in f['opcionales']] or None,
                    }
                } for f in self.facturas]
            })

        result = ret['FECAESolicitarResult']
        if 'FeCabResp' in result:
            fecabresp = result['FeCabResp']
            for i, fedetresp in enumerate(result['FeDetResp']):
                fedetresp = fedetresp['FECAEDetResponse']
                f = self.facturas[i]
                # actualizar los campos devueltos por AFIP para cada comp.
                f["resultado"] = fedetresp['Resultado']
                f["cae"] = fedetresp['CAE'] and str(fedetresp['CAE']) or ""
                f["emision_tipo"] = 'CAE'
                f["fch_venc_cae"] = fedetresp['CAEFchVto']
                f["obs"] = [
                    {'code': obs['Obs']['Code'], 'msg': obs['Obs']['Msg']}
                    for obs in fedetresp.get('Observaciones', [])
                ]
                # sanity checks:
                assert str(f["fecha_cbte"]) == str(fedetresp['CbteFch'])
                assert str(f["cbt_desde"]) == str(fedetresp['CbteDesde'])
                assert str(f["cbt_hasta"]) == str(fedetresp['CbteHasta'])
                assert str(f["punto_vta"]) == str(fecabresp['PtoVta'])
                assert str(f["tipo_cbte"]) == str(fecabresp['CbteTipo'])
                assert str(f["tipo_doc"]) == str(fedetresp['DocTipo'])
                assert str(f["nro_doc"]) == str(fedetresp['DocNro'])
                assert str(f["concepto"]) == str(fedetresp['Concepto'])

            self.__analizar_errores(result)
            assert fecabresp['CantReg'] == len(self.facturas)
        return fecabresp['CantReg']

    # metodos auxiliares para soporte de multiples comprobantes por solicitud:

    def IniciarFacturasX(self):
        "Inicializa lista de facturas para Solicitar multiples CAE"
        self.facturas = []
        return True

    def AgregarFacturaX(self):
        "Agrega una factura a la lista para Solicitar multiples CAE"
        self.facturas.append(self.factura)
        return True

    def LeerFacturaX(self, i):
        "Activa internamente una factura para usar ObtenerCampoFactura"
        try:
            # obtengo la factura segun el indice en la lista:
            f = self.factura = self.facturas[i]
            # completar propiedades para retro-compatibilidad:
            self.FechaCbte = f['fecha_cbte']
            self.PuntoVenta = f['punto_vta']
            self.Vencimiento = f['fch_venc_cae']
            self.Resultado = f['resultado']
            self.CbtDesde = f['cbt_desde']
            self.CbtHasta = f['cbt_hasta']
            self.ImpTotal = str(f['imp_total'])
            self.ImpNeto = str(f.get('imp_neto'))
            self.ImptoLiq = self.ImpIVA = str(f.get('imp_iva'))
            self.ImpOpEx = str(f.get('imp_op_ex'))
            self.ImpTrib = str(f.get('imp_trib'))
            self.EmisionTipo = f['emision_tipo']
            if self.EmisionTipo == 'CAE':
                self.CAE = f['cae']
            elif self.EmisionTipo == 'CAEA':
                self.CAEA = f['caea']
            # Obs:
            self.Observaciones = []
            for obs in f.get('obs', []):
                self.Observaciones.append("%(code)s: %(msg)s" % (obs))
            self.Obs = '\n'.join(self.Observaciones)
            return True
        except Exception as e:
            print(e)
            return False

    # metodos para CAEA:

    @inicializar_y_capturar_excepciones
    def CAEASolicitar(self, periodo, orden):
        ret = self.client.FECAEASolicitar(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            Periodo=periodo,
            Orden=orden,
        )

        result = ret['FECAEASolicitarResult']
        self.__analizar_errores(result)

        if 'ResultGet' in result:
            result = result['ResultGet']
            if 'CAEA' in result:
                self.CAEA = result['CAEA']
                self.Periodo = result['Periodo']
                self.Orden = result['Orden']
                self.FchVigDesde = result['FchVigDesde']
                self.FchVigHasta = result['FchVigHasta']
                self.FchTopeInf = result['FchTopeInf']
                self.FchProceso = result['FchProceso']
                # Obs (COMPGv28):
                for obs in result.get('Observaciones', []):
                    self.Observaciones.append(
                        "%(Code)s: %(Msg)s" % (obs['Obs'])
                    )
                self.Obs = '\n'.join(self.Observaciones)

        return self.CAEA and str(self.CAEA) or ''

    @inicializar_y_capturar_excepciones
    def CAEAConsultar(self, periodo, orden):
        "M�todo de consulta de CAEA"
        ret = self.client.FECAEAConsultar(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            Periodo=periodo,
            Orden=orden,
            )

        result = ret['FECAEAConsultarResult']
        self.__analizar_errores(result)

        if 'ResultGet' in result:
            result = result['ResultGet']
            if 'CAEA' in result:
                self.CAEA = result['CAEA']
                self.Periodo = result['Periodo']
                self.Orden = result['Orden']
                self.FchVigDesde = result['FchVigDesde']
                self.FchVigHasta = result['FchVigHasta']
                self.FchTopeInf = result['FchTopeInf']
                self.FchProceso = result['FchProceso']

        return self.CAEA and str(self.CAEA) or ''

    @inicializar_y_capturar_excepciones
    def CAEARegInformativo(self):
        "M�todo para informar comprobantes emitidos con CAEA"
        f = self.factura
        ret = self.client.FECAEARegInformativo(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            FeCAEARegInfReq={
                'FeCabReq': {
                    'CantReg': 1,
                    'PtoVta': f['punto_vta'],
                    'CbteTipo': f['tipo_cbte']},
                'FeDetReq': [{'FECAEADetRequest': {
                    'Concepto': f['concepto'],
                    'DocTipo': f['tipo_doc'],
                    'DocNro': f['nro_doc'],
                    'CbteDesde': f['cbt_desde'],
                    'CbteHasta': f['cbt_hasta'],
                    'CbteFch': f['fecha_cbte'],
                    'ImpTotal': f['imp_total'],
                    'ImpTotConc': f['imp_tot_conc'],
                    'ImpNeto': f['imp_neto'],
                    'ImpOpEx': f['imp_op_ex'],
                    'ImpTrib': f['imp_trib'],
                    'ImpIVA': f['imp_iva'],
                    # Fechas solo se informan si Concepto in (2,3)
                    'FchServDesde': f.get('fecha_serv_desde'),
                    'FchServHasta': f.get('fecha_serv_hasta'),
                    'FchVtoPago': f['fecha_venc_pago'],
                    'MonId': f['moneda_id'],
                    'MonCotiz': f['moneda_ctz'],
                    'CbtesAsoc': [
                        {'CbteAsoc': {
                            'Tipo': cbte_asoc['tipo'],
                            'PtoVta': cbte_asoc['pto_vta'],
                            'Nro': cbte_asoc['nro'],
                            'Cuit': cbte_asoc.get('cuit'),
                        }
                        }
                        for cbte_asoc in f['cbtes_asoc']
                    ] if f['cbtes_asoc'] else None,
                    'Tributos': [
                        {
                            'Tributo': {
                                'Id': tributo['tributo_id'],
                                'Desc': tributo['desc'],
                                'BaseImp': tributo['base_imp'],
                                'Alic': tributo['alic'],
                                'Importe': tributo['importe'],
                            }
                        } for tributo in f['tributos']
                    ] if f['tributos'] else None,
                    'Iva': [
                        {'AlicIva': {
                            'Id': iva['iva_id'],
                            'BaseImp': iva['base_imp'],
                            'Importe': iva['importe'],
                            }} for iva in f['iva']
                    ] if f['iva'] else None,
                    'CAEA': f['caea'],
                    }
                }]
            })

        result = ret['FECAEARegInformativoResult']
        if 'FeCabResp' in result:
            fecabresp = result['FeCabResp']
            fedetresp = result['FeDetResp'][0]['FECAEADetResponse']

            # Reprocesar en caso de error (recuperar CAE emitido anteriormente)
            if self.Reprocesar and 'Errors' in result:
                for error in result['Errors']:
                    err_code = str(error['Err']['Code'])
                    if fedetresp['Resultado'] == 'R' and err_code == '703':
                        caea = self.CompConsultar(
                            f['tipo_cbte'],
                            f['punto_vta'],
                            f['cbt_desde'],
                            reproceso=True
                        )
                        if caea and self.EmisionTipo == 'CAE':
                            self.Reproceso = 'S'
                            return caea
                        self.Reproceso = 'N'

            self.Resultado = fecabresp['Resultado']
            # Obs:
            for obs in fedetresp.get('Observaciones', []):
                self.Observaciones.append("%(Code)s: %(Msg)s" % (obs['Obs']))
            self.Obs = '\n'.join(self.Observaciones)
            self.CAEA = fedetresp['CAEA'] and str(fedetresp['CAEA']) or ""
            self.EmisionTipo = 'CAEA'
            self.FechaCbte = fedetresp['CbteFch']  # .strftime("%Y/%m/%d")
            self.CbteNro = fedetresp['CbteHasta']  # 1L
            self.PuntoVenta = fecabresp['PtoVta']  # 4000
            self.CbtDesde = fedetresp['CbteDesde']
            self.CbtHasta = fedetresp['CbteHasta']
            self.__analizar_errores(result)
        return self.CAEA

    @inicializar_y_capturar_excepciones
    def CAEASinMovimientoInformar(self, punto_vta, caea):
        "M�todo  para informar CAEA sin movimiento"
        ret = self.client.FECAEASinMovimientoInformar(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            PtoVta=punto_vta,
            CAEA=caea,
            )

        result = ret['FECAEASinMovimientoInformarResult']
        self.__analizar_errores(result)

        if 'CAEA' in result:
            self.CAEA = result['CAEA']
        if 'FchProceso' in result:
            self.FchProceso = result['FchProceso']
        if 'Resultado' in result:
            self.Resultado = result['Resultado']
            self.PuntoVenta = result['PtoVta']  # 4000

        return self.Resultado or ''

    @inicializar_y_capturar_excepciones
    def ParamGetTiposCbte(self, sep="|"):
        """Recuperador de valores referenciales de c�digos de Tipos de
        Comprobantes"""
        log.warn("### Calling Recuperador de valores referenciales de \
        códigos de Tipos de comprobante (FEParamGetTiposCbte)")
        ret = self.client.FEParamGetTiposCbte(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposCbteResult']
        log.warn("FEParamGetTiposCbte Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['CbteTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposConcepto(self, sep="|"):
        "Recuperador de valores referenciales de c�digos de Tipos de Conceptos"
        log.warn("### Calling Recuperador de valores referenciales de \
        códigos de Tipos de Conceptos FEParamGetTiposConcepto)")
        ret = self.client.FEParamGetTiposConcepto(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposConceptoResult']
        log.warn("FEParamGetTiposConcepto Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['ConceptoTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposDoc(self, sep="|"):
        """Recuperador de valores referenciales de c�digos de Tipos de
        Documentos"""
        log.warn("### Calling Recuperador de valores referenciales de \
        tipos de Documentos (FEParamGetTiposDoc)")
        ret = self.client.FEParamGetTiposDoc(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposDocResult']
        log.warn("FEParamGetTiposDoc Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['DocTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposIva(self, sep="|"):
        "Recuperador de valores referenciales de c�digos de Tipos de Al�cuotas"
        log.warn("### Calling Recuperador de valores referenciales de \
        codigos de Tipos de Alicuotas (FEParamGetTiposIva)")
        ret = self.client.FEParamGetTiposIva(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposIvaResult']
        log.warn("FEParamGetTiposIva Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['IvaTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposMonedas(self, sep="|"):
        "Recuperador de valores referenciales de c�digos de Monedas"
        log.warn("### Calling Recuperador de valores referenciales de \
        codigos de Tipos de Monedas (FEParamGetTiposMonedas)")
        ret = self.client.FEParamGetTiposMonedas(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposMonedasResult']
        log.warn("FEParamGetTiposMonedas Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['Moneda']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposOpcional(self, sep="|"):
        """Recuperador de valores referenciales de c�digos de Tipos de datos
        opcionales"""
        log.warn("### Calling Recuperador de valores referenciales de \
        codigos de Tipos de datos Opcionales (FEParamGetTiposOpcional)")
        ret = self.client.FEParamGetTiposOpcional(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposOpcionalResult']
        log.warn("FEParamGetTiposOpcional Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['OpcionalTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposTributos(self, sep="|"):
        """Recuperador de valores referenciales de c�digos de Tipos de
        Tributos Este m�todo permite consultar los tipos de tributos
        habilitados en este WS"""
        log.warn("### Calling Recuperador de valores referenciales de \
        codigos de Tipos de Tributos (FEParamGetTiposTributos)")
        ret = self.client.FEParamGetTiposTributos(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposTributosResult']
        log.warn("FEParamGetTiposTributos Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s\t%(FchDesde)s\t%(FchHasta)s" % (
                    p['TributoTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetTiposPaises(self, sep="|"):
        """Recuperador de valores referenciales de c�digos de Paises
        Este m�todo permite consultar los tipos de tributos habilitados en
        este WS"""
        log.warn("### Calling Recuperador de valores referenciales de \
                 codigos de Paises")
        ret = self.client.FEParamGetTiposPaises(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret['FEParamGetTiposPaisesResult']
        log.warn("FEParamGetTiposPaises Result: " + str(res))
        return [
            (
                "%(Id)s\t%(Desc)s" % (
                    p['PaisTipo']
                )
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]

    @inicializar_y_capturar_excepciones
    def ParamGetCotizacion(self, moneda_id):
        "Recuperador de cotizaci�n de moneda"
        log.warn("### Recuperador de cotizaci�n de moneda \
                (FEParamGetCotizacion)")
        ret = self.client.FEParamGetCotizacion(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
            MonId=moneda_id,
        )
        self.__analizar_errores(ret)
        res = ret['FEParamGetCotizacionResult']['ResultGet']
        log.warn("FEParamGetCotizacion Result: " + str(res))
        return str(res.get('MonCotiz', ""))

    @inicializar_y_capturar_excepciones
    def ParamGetPtosVenta(self, sep="|"):
        "Recuperador de valores referenciales Puntos de Venta registrados"
        log.warn("### Calling Recuperador de los puntos de venta asignados a \
        Facturaci�n Electronica que soporten CAE y CAEA via Web Services \
        (FEParamGetPtosVenta)")
        ret = self.client.FEParamGetPtosVenta(
            Auth={
                'Token': self.Token.decode('utf8'),
                'Sign': self.Sign.decode('utf8'),
                'Cuit': self.Cuit
            },
        )
        res = ret.get('FEParamGetPtosVentaResult', {})
        log.warn("FEParamGetPtosVenta Result: " + str(res))
        return [
            (
                """%(Nro)s\tEmisionTipo:%(EmisionTipo)s\tBloqueado:
                %(Bloqueado)s\tFchBaja:%(FchBaja)s""" % (p['PtoVenta'])
            ).replace("\t", sep) for p in res.get('ResultGet', [])
        ]


def p_assert_eq(a, b):
    print(a, a == b and '==' or '!=', b)


def main():
    "Funci�n principal de pruebas (obtener CAE)"
    import time

    DEBUG = '--debug' in sys.argv

    if DEBUG:
        from py3simplesoap.client import __version__ as soapver
        print("py3simplesoap.__version__ = ", soapver)

    wsfev1 = WSFEv1()
    wsfev1.LanzarExcepciones = True

    cache = None
    if "--prod" in sys.argv:
        wsdl = "https://servicios1.afip.gov.ar/wsfev1/service.asmx?WSDL"
    else:
        wsdl = WSDL
    proxy = ""
    wrapper = ""  # "pycurl"
    # cacert = True  # geotrust.crt"

    ok = wsfev1.Conectar(cache, wsdl, proxy, wrapper, cacert='')

    if not ok:
        raise RuntimeError(wsfev1.Excepcion)

    if DEBUG:
        print("LOG: ", wsfev1.DebugLog())

    if "--dummy" in sys.argv:
        print(wsfev1.client.help("FEDummy"))
        wsfev1.Dummy()
        print("AppServerStatus", wsfev1.AppServerStatus)
        print("DbServerStatus", wsfev1.DbServerStatus)
        print("AuthServerStatus", wsfev1.AuthServerStatus)
        sys.exit(0)

    # obteniendo el TA para pruebas
    from .wsaa import WSAA
    ta = WSAA().Autenticar("wsfe", "reingart.crt", "reingart.key", debug=True)
    wsfev1.SetTicketAcceso(ta)
    wsfev1.Cuit = "20267565393"

    if "--prueba" in sys.argv:
        print(wsfev1.client.help("FECAESolicitar").encode("latin1"))

        if '--usados' in sys.argv:
            tipo_cbte = 49
            concepto = 1
        elif '--fce' in sys.argv:
            tipo_cbte = 203
            concepto = 1
        else:
            tipo_cbte = 3
            concepto = 3 if ('--rg4109' not in sys.argv) else 1
        punto_vta = 4001
        cbte_nro = int(wsfev1.CompUltimoAutorizado(tipo_cbte, punto_vta) or 0)
        fecha = datetime.datetime.now().strftime("%Y%m%d")
        tipo_doc = 80 if '--usados' not in sys.argv else 30
        nro_doc = "30500010912"
        cbt_desde = cbte_nro + 1
        cbt_hasta = cbte_nro + 1
        imp_total = "222.00"
        imp_tot_conc = "0.00"
        imp_neto = "200.00"
        imp_iva = "21.00"
        imp_trib = "1.00"
        imp_op_ex = "0.00"
        fecha_cbte = fecha
        fecha_venc_pago = fecha_serv_desde = fecha_serv_hasta = None
        # Fechas del per�odo del servicio facturado y vencimiento de pago:
        if concepto > 1:
            fecha_venc_pago = fecha
            fecha_serv_desde = fecha
            fecha_serv_hasta = fecha
        elif '--fce' in sys.argv:
            # obligatorio en Factura de Crédito Electrónica MiPyMEs (FCE):
            fecha_venc_pago = fecha
        moneda_id = 'PES'
        moneda_ctz = '1.000'

        # inicializar prueba de multiples comprobantes por solicitud
        if "--multiple" in sys.argv:
            wsfev1.IniciarFacturasX()
            reg_x_req = wsfev1.CompTotXRequest()    # cant max. comprobantes
        else:
            reg_x_req = 1                           # un solo comprobante

        for i in range(reg_x_req):

            wsfev1.CrearFactura(
                concepto, tipo_doc, nro_doc,
                tipo_cbte, punto_vta, cbt_desde + i, cbt_hasta + i,
                imp_total, imp_tot_conc, imp_neto,
                imp_iva, imp_trib, imp_op_ex, fecha_cbte, fecha_venc_pago,
                fecha_serv_desde, fecha_serv_hasta,  # --
                moneda_id, moneda_ctz
            )

            if '--caea' in sys.argv:
                periodo = datetime.datetime.today().strftime("%Y%M")
                orden = 1 if datetime.datetime.today().day < 15 else 2
                caea = wsfev1.CAEAConsultar(periodo, orden)
                wsfev1.EstablecerCampoFactura("caea", caea)

             # comprobantes asociados (notas de crédito / débito)
            if tipo_cbte in (2, 3, 7, 8, 12, 13, 203, 208, 213):
                tipo = 201 if tipo_cbte in (203, 208, 213) else 3
                pto_vta = 4001
                nro = 1
                cuit = "20267565393"
                # obligatorio en Factura de Crédito Electrónica MiPyMEs (FCE):
                fecha_cbte = fecha if tipo_cbte in (203, 208, 213) else None
                wsfev1.AgregarCmpAsoc(tipo, pto_vta, nro, cuit, fecha_cbte)

            # otros tributos:
            tributo_id = 99
            desc = 'Impuesto Municipal Matanza'
            base_imp = None
            alic = None
            importe = 1
            wsfev1.AgregarTributo(tributo_id, desc, base_imp, alic, importe)

            # subtotales por alicuota de IVA:
            iva_id = 3  # 0%
            base_imp = 100
            importe = 0
            wsfev1.AgregarIva(iva_id, base_imp, importe)

            # subtotales por alicuota de IVA:
            iva_id = 5  # 21%
            base_imp = 100
            importe = 21
            wsfev1.AgregarIva(iva_id, base_imp, importe)

            # datos opcionales para proyectos promovidos:
            if '--proyectos' in sys.argv:
                wsfev1.AgregarOpcional(2, "1234")  # identificador del proyecto
            # datos opcionales para RG Bienes Usados 3411 (del vendedor):
            if '--usados' in sys.argv:
                wsfev1.AgregarOpcional(91, "Juan Perez")   # Nombre y Apellido
                wsfev1.AgregarOpcional(92, "200")          # Nacionalidad
                wsfev1.AgregarOpcional(93, "Balcarce 50")  # Domicilio
            # datos opcionales para RG 3668 Impuesto al Valor Agregado -
            # Art.12:
            if '--rg3668' in sys.argv:
                wsfev1.AgregarOpcional(5, "02")             # IVA Excepciones
                wsfev1.AgregarOpcional(61, "80")            # Firmante Doc Tipo
                wsfev1.AgregarOpcional(62, "20267565393")   # Firmante Doc Nro
                wsfev1.AgregarOpcional(7, "01")         # Car�cter del Firmante
            # datos opcionales para RG 4004-E Alquiler de inmuebles (Ganancias)
            if '--rg4004' in sys.argv:
                wsfev1.AgregarOpcional(17, "1")              # Intermediario
                wsfev1.AgregarOpcional(1801, "30500010912")  # CUIT Propietario
                wsfev1.AgregarOpcional(1802, "BNA")          # Nombr e Titular
            # datos de compradores RG 4109-E bienes muebles registrables (%)
            if '--rg4109' in sys.argv:
                wsfev1.AgregarComprador(80, "30500010912", 99.99)
                wsfev1.AgregarComprador(80, "30999032083", 0.01)
                
            # datos de Factura de Crédito Electrónica MiPyMEs (FCE):
            if '--fce' in sys.argv:  # 0110001900000000000000
                wsfev1.AgregarOpcional(2101, "2850590940090418135201")  # CBU
                wsfev1.AgregarOpcional(2102, "pyafipws")               # alias
                if tipo_cbte in (203, 208, 213):
                    wsfev1.AgregarOpcional(22, "S")  # Anulación

            # agregar la factura creada internamente para solicitud m�ltiple:
            if "--multiple" in sys.argv:
                wsfev1.AgregarFacturaX()

        t0 = time.time()
        if '--caea' not in sys.argv:
            if "--multiple" not in sys.argv:
                wsfev1.CAESolicitar()
            else:
                cant = wsfev1.CAESolicitarX()
                print("Cantidad de comprobantes procesados:", cant)
        else:
            wsfev1.CAEARegInformativo()
        t1 = time.time()

        # revisar los resultados:
        for i in range(reg_x_req):
            if "--multiple" in sys.argv:
                print("Analizando respuesta para factura indice: ", i)
                ok = wsfev1.LeerFacturaX(i)
            print("Nro. Cbte. desde-hasta", wsfev1.CbtDesde, wsfev1.CbtHasta)
            print("Resultado", wsfev1.Resultado)
            print("Reproceso", wsfev1.Reproceso)
            print("CAE", wsfev1.CAE)
            print("Vencimiento", wsfev1.Vencimiento)
            print("Observaciones", wsfev1.Obs)

        if DEBUG:
            print("t0", t0)
            print("t1", t1)
            print("lapso", t1-t0)
            open("xmlrequest.xml", "wb").write(wsfev1.XmlRequest)
            open("xmlresponse.xml", "wb").write(wsfev1.XmlResponse)

        if "--multiple" not in sys.argv:
            wsfev1.AnalizarXml("XmlResponse")
            p_assert_eq(wsfev1.ObtenerTagXml('CAE'), str(wsfev1.CAE))
            p_assert_eq(wsfev1.ObtenerTagXml('Concepto'), '2')
            p_assert_eq(wsfev1.ObtenerTagXml('Obs', 0, 'Code'), "10017")
            print(wsfev1.ObtenerTagXml('Obs', 0, 'Msg'))

        if "--reprocesar" in sys.argv:
            print("reprocesando....")
            wsfev1.Reproceso = True
            cae = wsfev1.CAE
            wsfev1.CAESolicitar()
            assert cae == wsfev1.CAE
            assert wsfev1.Reproceso == "S"

        if "--consultar" in sys.argv:
            cae = wsfev1.CAE
            cae2 = wsfev1.CompConsultar(tipo_cbte, punto_vta, cbt_desde)
            p_assert_eq(cae, cae2)
            # comparar datos del encabezado
            p_assert_eq(wsfev1.ObtenerCampoFactura('cae'), str(wsfev1.CAE))
            p_assert_eq(wsfev1.ObtenerCampoFactura('nro_doc'), int(nro_doc))
            p_assert_eq(
                wsfev1.ObtenerCampoFactura('imp_total'), float(imp_total)
            )
            # comparar primer alicuota de IVA
            p_assert_eq(wsfev1.ObtenerCampoFactura('iva', 0, 'importe'), 21)
            # comparar primer tributo
            p_assert_eq(
                wsfev1.ObtenerCampoFactura('tributos', 0, 'importe'), 1
            )
            # comparar primer opcional
            if '--rg3668' in sys.argv:
                p_assert_eq(
                    wsfev1.ObtenerCampoFactura('opcionales', 0, 'valor'),
                    "02"
                )
            # comparar primer observacion de AFIP
            p_assert_eq(wsfev1.ObtenerCampoFactura('obs', 0, 'code'), 10017)
            # pruebo la segunda observacion inexistente
            p_assert_eq(wsfev1.ObtenerCampoFactura('obs', 1, 'code'), None)
            p_assert_eq(wsfev1.Excepcion, "El campo 1 solicitado no existe")

    if "--get" in sys.argv:
        tipo_cbte = 2
        punto_vta = 4001
        cbte_nro = wsfev1.CompUltimoAutorizado(tipo_cbte, punto_vta)

        wsfev1.CompConsultar(tipo_cbte, punto_vta, cbte_nro)

        print("FechaCbte = ", wsfev1.FechaCbte)
        print("CbteNro = ", wsfev1.CbteNro)
        print("PuntoVenta = ", wsfev1.PuntoVenta)
        print("ImpTotal =", wsfev1.ImpTotal)
        print("CAE = ", wsfev1.CAE)
        print("Vencimiento = ", wsfev1.Vencimiento)
        print("EmisionTipo = ", wsfev1.EmisionTipo)

        wsfev1.AnalizarXml("XmlResponse")
        p_assert_eq(wsfev1.ObtenerTagXml('CodAutorizacion'), str(wsfev1.CAE))
        p_assert_eq(wsfev1.ObtenerTagXml('CbteFch'), wsfev1.FechaCbte)
        p_assert_eq(wsfev1.ObtenerTagXml('MonId'), "PES")
        p_assert_eq(wsfev1.ObtenerTagXml('MonCotiz'), "1")
        p_assert_eq(wsfev1.ObtenerTagXml('DocTipo'), "80")
        p_assert_eq(wsfev1.ObtenerTagXml('DocNro'), "30500010912")

    if "--parametros" in sys.argv:
        import codecs
        import locale
        # import traceback
        if sys.stdout.encoding is None:
            sys.stdout = codecs.getwriter(
                locale.getpreferredencoding()
            )(sys.stdout, "replace")
            sys.stderr = codecs.getwriter(
                locale.getpreferredencoding()
            )(sys.stderr, "replace")

        print('\n'.join(wsfev1.ParamGetTiposDoc()))
        print("=== Tipos de Comprobante ===")
        print('\n'.join(wsfev1.ParamGetTiposCbte()))
        print("=== Tipos de Concepto ===")
        print('\n'.join(wsfev1.ParamGetTiposConcepto()))
        print("=== Tipos de Documento ===")
        print('\n'.join(wsfev1.ParamGetTiposDoc()))
        print("=== Alicuotas de IVA ===")
        print('\n'.join(wsfev1.ParamGetTiposIva()))
        print("=== Monedas ===")
        print('\n'.join(wsfev1.ParamGetTiposMonedas()))
        print("=== Tipos de datos opcionales ===")
        print('\n'.join(wsfev1.ParamGetTiposOpcional()))
        print("=== Tipos de Tributo ===")
        print('\n'.join(wsfev1.ParamGetTiposTributos()))
        print("=== Tipos de Paises ===")
        print('\n'.join(wsfev1.ParamGetTiposPaises()))
        print("=== Puntos de Venta ===")
        print('\n'.join(wsfev1.ParamGetPtosVenta()))

    if "--cotizacion" in sys.argv:
        print(wsfev1.ParamGetCotizacion('DOL'))

    if "--comptox" in sys.argv:
        print(wsfev1.CompTotXRequest())

    if "--ptosventa" in sys.argv:
        print(wsfev1.ParamGetPtosVenta())

    if "--solicitar-caea" in sys.argv:
        periodo = sys.argv[sys.argv.index("--solicitar-caea")+1]
        orden = sys.argv[sys.argv.index("--solicitar-caea")+2]

        if DEBUG:
            print(
                "Solicitando CAEA para periodo %s orden %s" % (periodo, orden)
            )

        caea = wsfev1.CAEASolicitar(periodo, orden)
        print("CAEA:", caea)

        if wsfev1.Observaciones:
            print("Observaciones:")
            for obs in wsfev1.Observaciones:
                print(obs)

        if wsfev1.Errores:
            print("Errores:")
            for error in wsfev1.Errores:
                print(error)

        if DEBUG:
            print("periodo:", wsfev1.Periodo)
            print("orden:", wsfev1.Orden)
            print("fch_vig_desde:", wsfev1.FchVigDesde)
            print("fch_vig_hasta:", wsfev1.FchVigHasta)
            print("fch_tope_inf:", wsfev1.FchTopeInf)
            print("fch_proceso:", wsfev1.FchProceso)

        if not caea:
            print('Consultando CAEA')
            caea = wsfev1.CAEAConsultar(periodo, orden)
            print("CAEA:", caea)
            if wsfev1.Errores:
                print("Errores:")
                for error in wsfev1.Errores:
                    print(error)

    if "--sinmovimiento-caea" in sys.argv:
        punto_vta = sys.argv[sys.argv.index("--sinmovimiento-caea")+1]
        caea = sys.argv[sys.argv.index("--sinmovimiento-caea")+2]

        if DEBUG:
            print(
                "Informando Punto Venta %s CAEA %s SIN MOVIMIENTO" % (
                    punto_vta, caea
                )
            )

        resultado = wsfev1.CAEASinMovimientoInformar(punto_vta, caea)
        print("Resultado:", resultado)
        print("fch_proceso:", wsfev1.FchProceso)

        if wsfev1.Errores:
            print("Errores:")
            for error in wsfev1.Errores:
                print(error)


# busco el directorio de instalaci�n (global para que no cambie si usan otra
# dll)
INSTALL_DIR = WSFEv1.InstallDir = get_install_dir()


if __name__ == '__main__':

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import pythoncom
        if TYPELIB:
            if '--register' in sys.argv:
                tlb = os.path.abspath(
                    os.path.join(INSTALL_DIR, "typelib", "wsfev1.tlb")
                )
                print("Registering %s" % (tlb,))
                tli = pythoncom.LoadTypeLib(tlb)
                pythoncom.RegisterTypeLib(tli, tlb)
            elif '--unregister' in sys.argv:
                k = WSFEv1
                pythoncom.UnRegisterTypeLib(
                    k._typelib_guid_,
                    k._typelib_version_[0],
                    k._typelib_version_[1],
                    0,
                    pythoncom.SYS_WIN32
                )
                print("Unregistered typelib")
        import win32com.server.register
        # print("_reg_class_spec_", WSFEv1._reg_class_spec_)
        win32com.server.register.UseCommandLine(WSFEv1)
    elif "/Automate" in sys.argv:
        # MS seems to like /automate to run the class factories.
        import win32com.server.localserver
        # win32com.server.localserver.main()
        # start the server.
        win32com.server.localserver.serve([WSFEv1._reg_clsid_])
    else:
        main()



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"M�dulo para generar c�digos de barra en Entrelazado 2 de 5 (I25)"

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2011 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.02d"

import os
import sys
import traceback
from PIL import Image, ImageFont, ImageDraw


class PyI25:
    "Interfaz para generar PDF de Factura Electr�nica"
    _public_methods_ = ['GenerarImagen',
                        'DigitoVerificadorModulo10'
                        ]
    _public_attrs_ = ['Version', 'Excepcion', 'Traceback']

    _reg_progid_ = "PyI25"
    _reg_clsid_ = "{5E6989E8-F658-49FB-8C39-97C74BC67650}"

    def __init__(self):
        self.Version = __version__
        self.Exception = self.Traceback = ""

    def GenerarImagen(self, codigo, archivo="barras.png",
                      basewidth=3, width=None, height=30, extension="PNG"):
        "Generar una im�gen con el c�digo de barras Interleaved 2 of 5"
        # basado de:
        #  * http://www.fpdf.org/en/script/script67.php
        #  * http://code.activestate.com/recipes/426069/

        wide = basewidth
        narrow = basewidth / 3

        # c�digos ancho/angostos (wide/narrow) para los d�gitos
        bars = ("nnwwn", "wnnnw", "nwnnw", "wwnnn", "nnwnw", "wnwnn", "nwwnn",
                "nnnww", "wnnwn", "nwnwn", "nn", "wn")

        # agregar un 0 al principio si el n�mero de d�gitos es impar
        if len(codigo) % 2:
            codigo = "0" + codigo

        if not width:
            width = (len(codigo) * 3) * basewidth + (10 * narrow)
            print(width)
            # width = 380
        # crear una nueva im�gen
        im = Image.new("1", (width, height))

        # agregar c�digos de inicio y final
        codigo = "::" + codigo.lower() + ";:"  # A y Z en el original

        # crear un drawer
        draw = ImageDraw.Draw(im)

        # limpiar la im�gen
        draw.rectangle(((0, 0), (im.size[0], im.size[1])), fill=256)

        xpos = 0
        # dibujar los c�digos de barras
        for i in range(0, len(codigo), 2):
            # obtener el pr�ximo par de d�gitos
            bar = ord(codigo[i]) - ord("0")
            space = ord(codigo[i + 1]) - ord("0")
            # crear la sequencia barras (1er d�gito=barras, 2do=espacios)
            seq = ""
            for s in range(len(bars[bar])):
                seq = seq + bars[bar][s] + bars[space][s]

            for s in range(len(seq)):
                if seq[s] == "n":
                    width = narrow
                else:
                    width = wide

                # dibujar barras impares (las pares son espacios)
                if not s % 2:
                    draw.rectangle(((xpos, 0), (xpos+width-1, height)), fill=0)
                xpos = xpos + width

        im.save(archivo, extension.upper())
        return True

    def DigitoVerificadorModulo10(self, codigo):
        "Rutina para el c�lculo del d�gito verificador 'm�dulo 10'"
        # http://www.consejo.org.ar/Bib_elect/diciembre04_CT/documentos/rafip1702.htm
        # Etapa 1: comenzar desde la izquierda, sumar todos los caracteres
        # ubicados en las posiciones impares.
        codigo = codigo.strip()
        if not codigo or not codigo.isdigit():
            return ''
        etapa1 = sum([int(c) for i, c in enumerate(codigo) if not i % 2])
        # Etapa 2: multiplicar la suma obtenida en la etapa 1 por el n�mero 3
        etapa2 = etapa1 * 3
        # Etapa 3: comenzar desde la izquierda, sumar todos los caracteres
        # que est�n ubicados en las posiciones pares.
        etapa3 = sum([int(c) for i, c in enumerate(codigo) if i % 2])
        # Etapa 4: sumar los resultados obtenidos en las etapas 2 y 3.
        etapa4 = etapa2 + etapa3
        # Etapa 5: buscar el menor n�mero que sumado al resultado obtenido
        # en la etapa 4 d� un n�mero m�ltiplo de 10. Este ser� el valor del
        # d�gito verificador del m�dulo 10.
        digito = 10 - (etapa4 - (int(etapa4 / 10) * 10))
        if digito == 10:
            digito = 0
        return str(digito)


if __name__ == '__main__':

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(PyI25)
    elif "py2exe" in sys.argv:
        from distutils.core import setup
        from nsis import build_installer, Target
        import py2exe
        setup(
            name="PyI25",
            version=__version__,
            description="Interfaz PyAfipWs I25 %s",
            long_description=__doc__,
            author="Mariano Reingart",
            author_email="reingart@gmail.com",
            url="http://www.sistemasagiles.com.ar",
            license="GNU GPL v3",
            com_server=[
                {'modules': 'pyi25', 'create_exe': False, 'create_dll': True},
                ],
            console=['pyi25.py'],
            options={
                'py2exe': {
                    'includes': [],
                    'optimize': 2,
                    'excludes': [
                        "pywin",
                        "pywin.dialogs",
                        "pywin.dialogs.list",
                        "win32ui",
                        "distutils.core",
                        "py2exe",
                        "nsis"],
                    # 'skip_archive': True,
                }},
            data_files=[(".", ["licencia.txt"])],
            cmdclass={"py2exe": build_installer}
        )
    else:

        pyi25 = PyI25()

        if '--barras' in sys.argv:
            barras = sys.argv[sys.argv.index("--barras")+1]
        else:
            cuit = 20267565393
            tipo_cbte = 2
            punto_vta = 4001
            cae = 61203034739042
            fch_venc_cae = 20110529

            # codigo de barras de ejemplo:
            barras = '%11s%02d%04d%s%8s' % (
                cuit, tipo_cbte, punto_vta, cae, fch_venc_cae)

        if '--noverificador' not in sys.argv:
            barras = barras + pyi25.DigitoVerificadorModulo10(barras)

        if '--archivo' in sys.argv:
            archivo = sys.argv[sys.argv.index("--archivo")+1]
            extension = os.path.splitext(archivo)[1]
            extension = extension.upper()[1:]
            if extension == 'JPG':
                extension = 'JPEG'
        else:
            archivo = "prueba-cae-i25.png"
            extension = 'PNG'

        print("barras", barras)
        print("archivo", archivo)
        pyi25.GenerarImagen(barras, archivo, extension=extension)

        if '--mostrar' not in sys.argv:
            pass
        elif sys.platform == "linux2":
            os.system("eog ""%s""" % archivo)
        else:
            os.startfile(archivo)



from setuptools import setup

setup(
    name='py3afipws',
    packages=['py3afipws'],  # this must be the same as the name above
    version='0.10.26',
    description='pyafipws for python3',
    long_description='Factura Electrónica AFIP y otros servicios web',
    author='Mastercore',
    author_email='far@mastercore.net',
    # use the URL to the github repo
    url='https://github.com/odoo-mastercore/pyafipws',
    download_ur3='https://github.com/odoo-mastercore/py3afipws/tarball/0.10.11',
    keywords=['py3afipws', 'py3'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo para obtener un ticket de autorización del web service WSAA de AFIP"

# Basado en wsaa-client.php de Gerardo Fisanotti - DvSHyS/DiOPIN/AFIP -
# 13-apr-07
# Definir WSDL, CERT, PRIVATEKEY, PASSPHRASE, SERVICE, WSAAURL
# Devuelve TA.xml (ticket de autorización de WSAA)

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2008-2011 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "2.10f"

import hashlib
import datetime
# import email
import os
import sys
import time
import traceback
import warnings
import unicodedata
from .php import date
from py3simplesoap.client import SimpleXMLElement
from .utils import (
    inicializar_y_capturar_excepciones,
    BaseWS,
    get_install_dir,
    # exception_info,
    safe_console
)
from OpenSSL import crypto
from base64 import b64encode

import logging
log = logging.getLogger()

# Constantes (si se usa el script de linea de comandos)
# El WSDL correspondiente al WSAA
WSDL = "https://wsaahomo.afip.gov.ar/ws/services/LoginCms?wsdl"
CERT = "reingart.crt"        # El certificado X.509 obtenido de Seg. Inf.
PRIVATEKEY = "reingart.key"  # La clave privada del certificado CERT
PASSPHRASE = "xxxxxxx"  # La contraseña para firmar (si hay)
SERVICE = "wsfe"        # El nombre del web service al que se le pide el TA

# WSAAURL: la URL para acceder al WSAA, verificar http/https y wsaa/wsaahomo
WSAAURL = "https://wsaa.afip.gov.ar/ws/services/LoginCms"  # PRODUCCION!!!
# WSAAURL = "https://wsaahomo.afip.gov.ar/ws/services/LoginCms" # homologacion
SOAP_ACTION = 'http://ar.gov.afip.dif.facturaelectronica/'  # Revisar WSDL
SOAP_NS = "http://wsaa.view.sua.dvadac.desein.afip.gov"     # Revisar WSDL

# Verificación del web server remoto, necesario para verificar canal seguro
# CACERT = "conf/afip_ca_info.crt"  # WSAA CA Cert (Autoridades de Confiaza)
CACERT = ""  # WSAA CA Cert (Autoridades de Confiaza)

HOMO = False
TYPELIB = False
DEFAULT_TTL = 60*60*5       # five hours
DEBUG = False

# No debería ser necesario modificar nada despues de esta linea


def create_tra(service=SERVICE, ttl=2400):
    "Crear un Ticket de Requerimiento de Acceso (TRA)"
    tra = SimpleXMLElement(
        '<?xml version="1.0" encoding="UTF-8"?>'
        '<loginTicketRequest version="1.0">'
        '</loginTicketRequest>')
    tra.add_child('header')
    # El source es opcional. Si falta, toma la firma (recomendado).
    # tra.header.addChild('source','subject=...')
    # tra.header.addChild(
    #     'destination','cn=wsaahomo,o=afip,c=ar,serialNumber=CUIT 33693450239'
    # )
    tra.header.add_child('uniqueId', str(date('U')))
    tra.header.add_child('generationTime', str(date('c', date('U')-ttl)))
    tra.header.add_child('expirationTime', str(date('c', date('U')+ttl)))
    tra.add_child('service', service)
    return tra.as_xml().decode("utf8")


def sign_tra(tra, cert=CERT, privatekey=PRIVATEKEY, passphrase=""):
    "Firmar PKCS#7 el TRA y devolver CMS (recortando los headers SMIME)"
    log.warn("_______entrando en sign_tra_______")

    if True:
        try:
            tra = tra.encode("utf-8")
            try:
                pkey = crypto.load_privatekey(
                    crypto.FILETYPE_PEM, privatekey
                )
                sign_cert = crypto.load_certificate(
                    crypto.FILETYPE_PEM, cert
                )
            except Exception as e:
                log.warn('ERROR: %s' % str(e))
            PKCS7_NOSIGS = 0x4
            bio_in = crypto._new_mem_buf(tra)
            pkcs7 = crypto._lib.PKCS7_sign(
                sign_cert._x509, pkey._pkey, crypto._ffi.NULL, bio_in,
                PKCS7_NOSIGS
            )
            bio_out = crypto._new_mem_buf()
            crypto._lib.i2d_PKCS7_bio(bio_out, pkcs7)
            tra_signed = crypto._bio_to_string(bio_out)

            log.warn("tra_signed: " + str(tra_signed))
            log.warn("tra_signed: " + str(type(tra_signed)))
            return b64encode(tra_signed).decode("latin1")
        except Exception as e:
            log.warn("__out__ Exception: " + str(e))
            if isinstance(e, IOError) and e.errno == 2:
                warnings.warn(
                    "El ejecutable de OpenSSL no esta disponible en el PATH")
                raise


def call_wsaa(cms, location=WSAAURL, proxy=None, trace=False, cacert=None):
    "Llamar web service con CMS para obtener ticket de autorización (TA)"

    # creo la nueva clase
    wsaa = WSAA()
    try:
        wsaa.Conectar(proxy=proxy, wsdl=location, cache="", cacert=cacert)
        ta = wsaa.LoginCMS(cms)
        if not ta:
            raise RuntimeError(wsaa.Excepcion)
        else:
            return ta
    except Exception as e:
        print(e)
        raise


class WSAA(BaseWS):
    "Interfaz para el WebService de Autenticación y Autorización"
    _public_methods_ = [
        'CreateTRA', 'SignTRA', 'CallWSAA', 'LoginCMS', 'Conectar',
        'AnalizarXml', 'ObtenerTagXml', 'Expirado', 'Autenticar',
        'DebugLog', 'AnalizarCertificado',
        'CrearClavePrivada', 'CrearPedidoCertificado',
    ]
    _public_attrs_ = [
        'Token', 'Sign', 'ExpirationTime', 'Version',
        'XmlRequest', 'XmlResponse',
        'InstallDir', 'Traceback', 'Excepcion',
        'Identidad', 'Caducidad', 'Emisor',
        'SoapFault', 'LanzarExcepciones',
    ]
    _readonly_attrs_ = _public_attrs_[:-1]
    _reg_progid_ = "WSAA"
    _reg_clsid_ = "{6268820C-8900-4AE9-8A2D-F0A1EBD4CAC5}"

    if TYPELIB:
        _typelib_guid_ = '{30E9C94B-7385-4534-9A80-DF50FD169253}'
        _typelib_version_ = 2, 10
        _com_interfaces_ = ['IWSAA']

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    @inicializar_y_capturar_excepciones
    def CreateTRA(self, service="wsfe", ttl=2400):
        "Crear un Ticket de Requerimiento de Acceso (TRA)"
        return create_tra(service, ttl)

    @inicializar_y_capturar_excepciones
    def AnalizarCertificado(self, crt, binary=False):
        "Carga un certificado digital y extrae los campos más importantes"
        from M2Crypto import (
            BIO,
            # EVP,
            # RSA,
            X509
        )
        if binary:
            try:
                bio = BIO.MemoryBuffer(cert)
            except Exception as e:
                print(e)
                bio = BIO.MemoryBuffer(crt)
            x509 = X509.load_cert_bio(bio, X509.FORMAT_DER)
        else:
            if not crt.startswith("-----BEGIN CERTIFICATE-----"):
                crt = open(crt).read()
            bio = BIO.MemoryBuffer(crt)
            x509 = X509.load_cert_bio(bio, X509.FORMAT_PEM)
        if x509:
            self.Identidad = x509.get_subject().as_text()
            self.Caducidad = x509.get_not_after().get_datetime()
            self.Emisor = x509.get_issuer().as_text()
        return True

    @inicializar_y_capturar_excepciones
    def CrearClavePrivada(
        self, filename="privada.key", key_length=1024,
        pub_exponent=0x10001, passphrase=""
    ):
        "Crea una clave privada (private key)"
        from M2Crypto import RSA, EVP

        # only protect if passphrase was given (it will fail otherwise)
        callback = lambda *args, **kwarg: passphrase
        chiper = None if not passphrase else "aes_128_cbc"
        # create the RSA key pair (and save the result to a file):
        rsa_key_pair = RSA.gen_key(key_length, pub_exponent, callback)
        bio = BIO.MemoryBuffer()
        rsa_key_pair.save_key_bio(bio, chiper, callback)
        f = open(filename, "w")
        f.write(bio.read())
        f.close()
        # create a public key to sign the certificate request:
        self.pkey = EVP.PKey(md='sha1')
        self.pkey.assign_rsa(rsa_key_pair)

    @inicializar_y_capturar_excepciones
    def CrearPedidoCertificado(
        self, cuit="", empresa="", nombre="pyafipws",
        filename="empresa.csr"
    ):
        "Crear un certificate signing request (X509 CSR)"
        from M2Crypto import (
            RSA, EVP, X509
        )

        # create the certificate signing request (CSR):
        self.x509_req = X509.Request()

        # normalizar encoding (reemplazar acentos, eñe, etc.)
        if isinstance(empresa, str):
            empresa = unicodedata.normalize('NFKD', empresa)\
                    .encode('ASCII', 'ignore')
        if isinstance(nombre, str):
            nombre = unicodedata.normalize('NFKD', nombre)\
                    .encode('ASCII', 'ignore')

        # subjet: C=AR/O=[empresa]/CN=[nombre]/serialNumber=CUIT [nro_cuit]
        x509name = X509.X509_Name()
        # default OpenSSL parameters:
        kwargs = {"type": 0x1000 | 1, "len": -1, "loc": -1, "set": 0}
        x509name.add_entry_by_txt(field='C', entry='AR', **kwargs)
        x509name.add_entry_by_txt(field='O', entry=empresa, **kwargs)
        x509name.add_entry_by_txt(field='CN', entry=nombre, **kwargs)
        x509name.add_entry_by_txt(
            field='serialNumber', entry="CUIT %s" % cuit, **kwargs)
        self.x509_req.set_subject_name(x509name)

        # sign the request with the previously created key (CrearClavePrivada)
        self.x509_req.set_pubkey(pkey=self.pkey)
        self.x509_req.sign(pkey=self.pkey, md='sha256')
        # save the CSR result to a file:
        f = open(filename, "w")
        f.write(self.x509_req.as_pem())
        f.close()

    @inicializar_y_capturar_excepciones
    def SignTRA(self, tra, cert, privatekey, passphrase=""):
        "Firmar el TRA y devolver CMS"
        return sign_tra(
            tra,
            cert.encode('latin1'),
            privatekey.encode('latin1'),
            passphrase.encode("utf8")
        )

    @inicializar_y_capturar_excepciones
    def LoginCMS(self, cms):
        "Obtener ticket de autorización (TA)"
        log.warn("Entrando a LoginCMS")
        try:
            # log.warn("CMS: " + str(cms))
            results = self.client.loginCms(in0=str(cms))
        except Exception as e:
            log.warn("_______Chk1 ISSUE_______ " + str(e))
        ta_xml = results['loginCmsReturn']
        log.warn("_______Chk2_______")
        self.xml = ta = SimpleXMLElement(ta_xml)
        log.warn("_______Chk3_______")
        self.Token = str(ta.credentials.token)
        log.warn("_______Chk4_______")
        self.Sign = str(ta.credentials.sign)
        log.warn("_______Chk5_______")
        self.ExpirationTime = str(ta.header.expirationTime)
        return ta_xml

    def CallWSAA(self, cms, url="", proxy=None):
        "Obtener ticket de autorización (TA) -version retrocompatible-"
        self.Conectar("", url, proxy)
        ta_xml = self.LoginCMS(cms)
        if not ta_xml:
            raise RuntimeError(self.Excepcion)
        return ta_xml

    @inicializar_y_capturar_excepciones
    def Expirado(self, fecha=None):
        "Comprueba la fecha de expiración, devuelve si ha expirado"
        if not fecha:
            fecha = self.ObtenerTagXml('expirationTime')
        now = datetime.datetime.now()
        d = datetime.datetime.strptime(fecha[:19], '%Y-%m-%dT%H:%M:%S')
        return now > d

    def Autenticar(
        self, service, crt, key, wsdl=None, proxy=None, wrapper=None,
        cacert=None, cache=None, debug=False
    ):
        "Método unificado para obtener el ticket de acceso (cacheado)"

        self.LanzarExcepciones = True
        try:
            # sanity check: verificar las credenciales
            for filename in (crt, key):
                if not os.access(filename, os.R_OK):
                    raise RuntimeError("Imposible abrir %s\n" % filename)
            # creo el nombre para el archivo del TA (según credenciales y ws)
            hash = service + crt + key
            fn = "TA-%s.xml" % hashlib.md5(hash.encode("latin1")).hexdigest()
            if cache:
                fn = os.path.join(cache, fn)
            else:
                fn = os.path.join(self.InstallDir, "cache", fn)

            # leer el ticket de acceso (si fue previamente solicitado)
            if not os.path.exists(fn) or os.path.getsize(fn) == 0 or \
               os.path.getmtime(fn) + (DEFAULT_TTL) < time.time():
                # ticket de acceso (TA) vencido, crear un nuevo req. (TRA)
                if DEBUG:
                    print("Creando TRA...")
                tra = self.CreateTRA(service=service, ttl=DEFAULT_TTL)
                # firmarlo criptográficamente
                if DEBUG:
                    print("Frimando TRA...")
                cms = self.SignTRA(tra, crt, key)
                # concectar con el servicio web:
                if DEBUG:
                    print("Conectando a WSAA...")
                ok = self.Conectar(cache, wsdl, proxy, wrapper, cacert)
                if not ok or self.Excepcion:
                    raise RuntimeError(
                        "Fallo la conexión: %s" % self.Excepcion)
                # llamar al método remoto para solicitar el TA
                if DEBUG:
                    print("Llamando WSAA...")
                ta = self.LoginCMS(cms)
                if not ta:
                    raise RuntimeError(
                        "Ticket de acceso vacio: %s" % WSAA.Excepcion)
                # grabar el ticket de acceso para poder reutilizarlo luego
                if DEBUG:
                    print(("Grabando TA en %s..." % fn))
                try:
                    open(fn, "w").write(ta)
                except IOError as e:
                    print(e)
                    self.Excepcion = "Imposible grabar ticket de \
                            accesso: %s" % fn
            else:
                # leer el ticket de acceso del archivo en cache
                if DEBUG:
                    print(("Leyendo TA de %s..." % fn))
                ta = open(fn, "r").read()
            # analizar el ticket de acceso y extraer los datos relevantes
            self.AnalizarXml(xml=ta)
            self.Token = self.ObtenerTagXml("token")
            self.Sign = self.ObtenerTagXml("sign")
        except Exception as e:
            ta = ""
            if not self.Excepcion:
                # avoid encoding problem when reporting exceptions to the user:
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_info()[0],
                    sys.exc_info()[1])[0]
                self.Traceback = ""
            if DEBUG or debug:
                raise
        return ta


# busco el directorio de instalación (global para que no cambie si usan
# otra dll)
INSTALL_DIR = WSAA.InstallDir = get_install_dir()


if __name__ == "__main__":

    safe_console()

    if '--register' in sys.argv or '--unregister' in sys.argv:
        import pythoncom
        if TYPELIB:
            if '--register' in sys.argv:
                tlb = os.path.abspath(
                    os.path.join(INSTALL_DIR, "typelib", "wsaa.tlb"))
                print(("Registering %s" % (tlb,)))
                tli = pythoncom.LoadTypeLib(tlb)
                pythoncom.RegisterTypeLib(tli, tlb)
            elif '--unregister' in sys.argv:
                k = WSAA
                pythoncom.UnRegisterTypeLib(k._typelib_guid_,
                                            k._typelib_version_[0],
                                            k._typelib_version_[1],
                                            0,
                                            pythoncom.SYS_WIN32)
                print("Unregistered typelib")
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSAA)
    elif "/Automate" in sys.argv:
        # MS seems to like /automate to run the class factories.
        import win32com.server.localserver
        # win32com.server.localserver.main()
        # start the server.
        win32com.server.localserver.serve([WSAA._reg_clsid_])
    elif "--crear_pedido_cert" in sys.argv:
        # instanciar el helper y revisar los parámetros
        wsaa = WSAA()
        args = [arg for arg in sys.argv if not arg.startswith("--")]
        # obtengo el CUIT y lo normalizo:
        cuit = len(args) > 1 and args[1] or eval(input("Ingrese un CUIT: "))
        cuit = ''.join([c for c in cuit if c.isdigit()])
        nombre = len(args) > 2 and args[2] or "PyAfipWs"
        # consultar el padrón online de AFIP si no se especificó razón social:
        empresa = len(args) > 3 and args[3] or ""
        if not empresa:
            from .padron import PadronAFIP
            padron = PadronAFIP()
            ok = padron.Consultar(cuit)
            if ok and padron.denominacion:
                print(("Denominación según AFIP:", padron.denominacion))
                empresa = padron.denominacion
            else:
                print(
                    ("CUIT %s no encontrado: %s..." % (cuit, padron.Excepcion))
                )
                empresa = eval(input("Empresa: "))
        # generar los archivos (con fecha para no pisarlo)
        ts = datetime.datetime.now().strftime("%Y%m%d%M%S")
        clave_privada = "clave_privada_%s_%s.key" % (cuit, ts)
        pedido_cert = "pedido_cert_%s_%s.csr" % (cuit, ts)
        wsaa.CrearClavePrivada(clave_privada)
        wsaa.CrearPedidoCertificado(cuit, empresa, nombre, pedido_cert)
        print("Se crearon los archivos:")
        print(clave_privada)
        print(pedido_cert)
        # convertir a terminación de linea windows y abrir con bloc de notas
        if sys.platform == "win32":
            txt = open(pedido_cert + ".txt", "wb")
            for linea in open(pedido_cert, "r"):
                txt.write("%s\r\n" % linea)
            txt.close()
            os.startfile(pedido_cert + ".txt")
    else:
        # Leer argumentos desde la linea de comando (si no viene tomar default)
        args = [arg for arg in sys.argv if arg.startswith("--")]
        argv = [arg for arg in sys.argv if not arg.startswith("--")]
        crt = len(argv) > 1 and argv[1] or CERT
        key = len(argv) > 2 and argv[2] or PRIVATEKEY
        service = len(argv) > 3 and argv[3] or "wsfe"
        ttl = len(argv) > 4 and int(argv[4]) or 36000
        url = len(argv) > 5 and argv[5] or WSAAURL
        wrapper = len(argv) > 6 and argv[6] or None
        cacert = len(argv) > 7 and argv[7] or CACERT
        DEBUG = "--debug" in args

        log.warn(
            "Usando CRT=%s KEY=%s URL=%s SERVICE=%s TTL=%s" % (
                crt, key, url, service, ttl)
        )

        # creo el objeto para comunicarme con el ws
        wsaa = WSAA()
        wsaa.LanzarExcepciones = True

        log.warn("WSAA Version %s %s" % (WSAA.Version, HOMO))

        if '--proxy' in args:
            proxy = sys.argv[sys.argv.index("--proxy")+1]
            log.warn("Usando PROXY:", proxy)
        else:
            proxy = None

        if '--analizar' in sys.argv:
            wsaa.AnalizarCertificado(crt)
            print((wsaa.Identidad))
            print((wsaa.Caducidad))
            print((wsaa.Emisor))

        ta = wsaa.Autenticar(service, crt, key, url, proxy, wrapper, cacert)
        if not ta:
            if DEBUG:
                log.warn(wsaa.Traceback)
            sys.exit("Excepcion: %s" % wsaa.Excepcion)

        else:
            print(ta)

        if wsaa.Excepcion:
            log.warn(wsaa.Excepcion)

        if DEBUG:
            print(("Source:", wsaa.ObtenerTagXml('source')))
            print(("UniqueID Time:", wsaa.ObtenerTagXml('uniqueId')))
            print(("Generation Time:", wsaa.ObtenerTagXml('generationTime')))
            print(("Expiration Time:", wsaa.ObtenerTagXml('expirationTime')))
            print(("Expiro?", wsaa.Expirado()))
            # import time; time.sleep(10)
            print("Expiro?", wsaa.Expirado())



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo para obtener c�digo de autorizaci�n electr�nico (CAE) del web service
WSBFEv1 de AFIP (Bonos Fiscales electronicos v1.1 - Factura Electr�nica RG)
a fin de gestionar los Bonos en la Secretar�a de Industria seg�n RG 2557
"""

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2013-2015 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.06d"

import datetime
import decimal
import os
import sys
from utils import inicializar_y_capturar_excepciones, BaseWS, get_install_dir

HOMO = False
LANZAR_EXCEPCIONES = True      # valor por defecto: True
WSDL = "https://wswhomo.afip.gov.ar/wsbfev1/service.asmx?WSDL"


class WSBFEv1(BaseWS):
    "Interfaz para el WebService de Bono Fiscal Electr�nico V1 (FE Bs. Capital)"
    _public_methods_ = [
        'CrearFactura', 'AgregarItem', 'Authorize', 'GetCMP',
        'GetParamMon', 'GetParamTipoCbte', 'GetParamUMed',
        'GetParamTipoIVA', 'GetParamNCM', 'GetParamZonas',
        'Dummy', 'Conectar', 'GetLastCMP', 'GetLastID',
        'GetParamCtz', 'LoadTestXML',
        'AnalizarXml', 'ObtenerTagXml', 'DebugLog',
        'SetParametros', 'SetTicketAcceso', 'GetParametro',
        'Dummy', 'Conectar', 'SetTicketAcceso'
    ]
    _public_attrs_ = [
        'Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'XmlRequest', 'XmlResponse', 'Version',
        'Resultado', 'Obs', 'Reproceso', 'FechaCAE',
        'CAE', 'Vencimiento', 'Eventos', 'ErrCode', 'ErrMsg', 'FchVencCAE',
        'Excepcion', 'LanzarExcepciones', 'Traceback', "InstallDir",
        'PuntoVenta', 'CbteNro', 'FechaCbte', 'ImpTotal', 'ImpNeto', 'ImptoLiq'
    ]

    _reg_progid_ = "WSBFEv1"
    _reg_clsid_ = "{EE4ABEE2-76DD-450F-880B-66710AE464D6}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologaci�n' or '')
    factura = None

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.CbteNro = self.FechaCbte = self.PuntoVenta = self.ImpTotal = None
        self.ImpNeto = self.ImptoLiq = None
        self.LanzarExcepciones = LANZAR_EXCEPCIONES
        self.InstallDir = INSTALL_DIR
        self.FechaCAE = self.FchVencCAE = ""   # retrocompatibilidad

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        if 'BFEErr' in ret:
            errores = [ret['BFEErr']]
            for error in errores:
                self.Errores.append("%s: %s" % (
                    error['ErrCode'],
                    error['ErrMsg'],
                    ))
            self.ErrCode = ' '.join([str(error['ErrCode']) for error in errores])
            self.ErrMsg = '\n'.join(self.Errores)
        if 'BFEEvents' in ret:
            events = [ret['BFEEvents']]
            self.Eventos = ['%s: %s' % (evt['EventCode'], evt.get('EventMsg',"")) for evt in events]

    def CrearFactura(self, tipo_doc=80, nro_doc=23111111113,
            zona=0, tipo_cbte=1, punto_vta=1, cbte_nro=0, fecha_cbte=None,
            imp_total=0.0, imp_neto=0.0, impto_liq=0.0,
            imp_tot_conc=0.0, impto_liq_rni=0.00, imp_op_ex=0.00,
            imp_perc=0.00, imp_iibb=0.00, imp_perc_mun=0.00, imp_internos=0.00,
            imp_moneda_id=0, imp_moneda_ctz=1.0, **kwargs):
        "Creo un objeto factura (interna)"
        # Creo una factura para bonos fiscales electr�nicos

        fact = {'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
                'cbte_nro': cbte_nro, 'fecha_cbte': fecha_cbte, 'zona': zona,
                'tipo_doc': tipo_doc, 'nro_doc':  nro_doc,
                'imp_total': imp_total, 'imp_neto': imp_neto,
                'impto_liq': impto_liq, 'impto_liq_rni': impto_liq_rni,
                'imp_op_ex': imp_op_ex, 'imp_tot_conc': imp_tot_conc,
                'imp_perc': imp_perc, 'imp_perc_mun': imp_perc_mun,
                'imp_iibb': imp_iibb, 'imp_internos': imp_internos,
                'imp_moneda_id': imp_moneda_id, 'imp_moneda_ctz': imp_moneda_ctz,
                'cbtes_asoc': [],
                'iva': [],
                'detalles': [],
            }
        self.factura = fact
        return True

    def AgregarItem(self, ncm, sec, ds, qty, umed, precio, bonif, iva_id, imp_total, **kwargs):
        "Agrego un item a una factura (interna)"
        ##ds = unicode(ds, "latin1") # convierto a latin1
        # Nota: no se calcula neto, iva, etc (deben venir calculados!)
        self.factura['detalles'].append({
                'ncm': ncm, 'sec': sec,
                'ds': ds,
                'qty': qty,
                'umed': umed,
                'precio': precio,
                'bonif': bonif,
                'iva_id': iva_id,
                'imp_total': imp_total,
                })
        return True

    @inicializar_y_capturar_excepciones
    def Authorize(self, id):
        "Autoriza la factura cargada en memoria"
        f = self.factura
        ret = self.client.BFEAuthorize(
            Auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit},
            Cmp={
                'Id': id,
                'Zona': f['zona'],
                'Fecha_cbte': f['fecha_cbte'],
                'Tipo_cbte': f['tipo_cbte'],
                'Punto_vta': f['punto_vta'],
                'Cbte_nro': f['cbte_nro'],
                'Tipo_doc': f['tipo_doc'], 'Nro_doc': f['nro_doc'],
                'Imp_moneda_Id': f['imp_moneda_id'],
                'Imp_moneda_ctz': f['imp_moneda_ctz'],
                'Imp_total': f['imp_total'],
                'Imp_tot_conc': f['imp_tot_conc'], 'Imp_op_ex': f['imp_op_ex'],
                'Imp_neto': f['imp_neto'], 'Impto_liq': f['impto_liq'],
                'Impto_liq_rni': f['impto_liq_rni'],
                'Imp_perc': f['imp_perc'], 'Imp_perc_mun': f['imp_perc_mun'],
                'Imp_iibb': f['imp_iibb'],
                'Imp_internos': f['imp_internos'],
                'Items': [
                    {'Item': {
                        'Pro_codigo_ncm': d['ncm'],
                        'Pro_codigo_sec': d['sec'],
                        'Pro_ds': d['ds'],
                        'Pro_qty': d['qty'],
                        'Pro_umed': d['umed'],
                        'Pro_precio_uni': d['precio'],
                        'Imp_bonif': d['bonif'],
                        'Imp_total': d['imp_total'],
                        'Iva_id': d['iva_id'],
                     }} for d in f['detalles']],
            })

        result = ret['BFEAuthorizeResult']
        self.__analizar_errores(result)
        if 'BFEResultAuth' in result:
            auth = result['BFEResultAuth']
            # Resultado: A: Aceptado, R: Rechazado
            self.Resultado = auth.get('Resultado', "")
            # Obs:
            self.Obs = auth.get('Obs', "")
            self.Reproceso = auth.get('Reproceso', "")
            self.CAE = auth.get('Cae', "")
            self.CbteNro  = auth.get('Fch_cbte', "")
            self.ImpTotal = str(auth.get('Imp_total', ''))
            self.ImptoLiq = str(auth.get('Impto_liq', ''))
            self.ImpNeto = str(auth.get('Imp_neto', ''))
            vto = str(auth.get('Fch_venc_Cae', ''))
            self.FchVencCAE = vto
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
            return self.CAE

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        result = self.client.BFEDummy()['BFEDummyResult']
        self.__analizar_errores(result)
        self.AppServerStatus = str(result['AppServer'])
        self.DbServerStatus = str(result['DbServer'])
        self.AuthServerStatus = str(result['AuthServer'])
        return True

    @inicializar_y_capturar_excepciones
    def GetCMP(self, tipo_cbte, punto_vta, cbte_nro):
        "Recuperar los datos completos de un comprobante ya autorizado"
        ret = self.client.BFEGetCMP(
            Auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit},
            Cmp={"Tipo_cbte": tipo_cbte,
             "Punto_vta": punto_vta, "Cbte_nro": cbte_nro})
        result = ret['BFEGetCMPResult']
        self.__analizar_errores(result)
        if 'BFEResultGet' in result:
            resultget = result['BFEResultGet']
            # Obs, cae y fecha cae
            if 'Cae' in resultget:
                self.Obs = resultget['Obs'] and resultget['Obs'].strip(" ") or ''
                self.CAE = resultget['Cae']
                vto = str(resultget['Fch_venc_Cae'])
                self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
                self.FechaCbte = resultget['Fecha_cbte_orig'] #.strftime("%Y/%m/%d")
                self.FechaCAE = resultget['Fecha_cbte_cae'] #.strftime("%Y/%m/%d")
                self.PuntoVenta = resultget['Punto_vta'] # 4000
                self.Resultado = resultget['Resultado']
                self.CbteNro =resultget['Cbte_nro']
                self.ImpTotal = resultget['Imp_total']
                self.ImptoLiq = resultget['Impto_liq']
                self.ImpNeto = resultget['Imp_neto']
            return self.CAE
        else:
            return 0

    @inicializar_y_capturar_excepciones
    def GetLastCMP(self, tipo_cbte, punto_vta):
        "Recuperar �ltimo n�mero de comprobante emitido"
        ret = self.client.BFEGetLast_CMP(
            Auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit,
                  "Tipo_cbte": tipo_cbte,
                  "Pto_venta": punto_vta,
            })
        result = ret['BFEGetLast_CMPResult']
        self.__analizar_errores(result)
        if 'BFEResult_LastCMP' in result:
            resultget = result['BFEResult_LastCMP']
            self.CbteNro =resultget.get('Cbte_nro')
            self.FechaCbte = resultget.get('Cbte_fecha') #.strftime("%Y/%m/%d")
            return self.CbteNro

    @inicializar_y_capturar_excepciones
    def GetLastID(self):
        "Recuperar �ltimo n�mero de transacci�n (ID)"
        ret = self.client.BFEGetLast_ID(
            Auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetLast_IDResult']
        self.__analizar_errores(result)
        if 'BFEResultGet' in result:
            resultget = result['BFEResultGet']
            return resultget.get('Id')

    @inicializar_y_capturar_excepciones
    def GetParamUMed(self):
        ret = self.client.BFEGetPARAM_UMed(
            auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetPARAM_UMedResult']
        self.__analizar_errores(result)

        umeds = [] # unidades de medida
        for u in result['BFEResultGet']:
            u = u['ClsBFEResponse_UMed']
            try:
                umed = {'id': u.get('Umed_Id'), 'ds': u.get('Umed_Ds'),
                        'vig_desde': u.get('Umed_vig_desde'),
                        'vig_hasta': u.get('Umed_vig_hasta')}
            except Exception as e:
                print(e)
                if u is None:
                    # <ClsFEXResponse_UMed xsi:nil="true"/> WTF!
                    umed = {'id':'', 'ds':'','vig_desde':'','vig_hasta':''}
                    # import pdb; pdb.set_trace()
                    # print(u)


            umeds.append(umed)
        return ['%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in umeds]

    @inicializar_y_capturar_excepciones
    def GetParamMon(self):
        ret = self.client.BFEGetPARAM_MON(
            auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetPARAM_MONResult']
        self.__analizar_errores(result)

        mons = [] # unidades de medida
        for m in result['BFEResultGet']:
            m = m['ClsBFEResponse_Mon']
            try:
                mon = {'id': m.get('Mon_Id'), 'ds': m.get('Mon_Ds'),
                        'vig_desde': m.get('Mon_vig_desde'),
                        'vig_hasta': m.get('Mon_vig_hasta')}
            except Exception as e:
                raise
                if m is None:
                    # <ClsFEXResponse_UMed xsi:nil="true"/> WTF!
                    mon = {'id':'', 'ds':'','vig_desde':'','vig_hasta':''}
                    # import pdb; pdb.set_trace()
                    # print(u)
            mons.append(mon)
        return ['%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in mons]

    @inicializar_y_capturar_excepciones
    def GetParamTipoIVA(self):
        "Recuperar lista de valores referenciales de tipos de IVA (al�cuotas)"
        ret = self.client.BFEGetPARAM_Tipo_IVA(
            auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetPARAM_Tipo_IVAResult']
        self.__analizar_errores(result)

        ivas = [] # tipos de iva
        for i in result['BFEResultGet']:
            i = i['ClsBFEResponse_Tipo_IVA']
            try:
                iva = {'id': i.get('IVA_Id'), 'ds': i.get('IVA_Ds'),
                        'vig_desde': i.get('IVA_vig_desde'),
                        'vig_hasta': i.get('IVA_vig_hasta')}
                ivas.append(iva)
            except Exception as e:
                pass
                raise
        return ['%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in ivas]

    def GetParamTipoCbte(self):
        "Recuperar lista de valores referenciales de Tipos de Comprobantes"
        ret = self.client.BFEGetPARAM_Tipo_Cbte(
            auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetPARAM_Tipo_CbteResult']
        self.__analizar_errores(result)

        tipos = [] # tipos de comprobantes
        for t in result['BFEResultGet']:
            t = t['ClsBFEResponse_Tipo_Cbte']
            try:
                tipo = {'id': t.get('Cbte_Id'), 'ds': t.get('Cbte_Ds'),
                        'vig_desde': t.get('Cbte_vig_desde'),
                        'vig_hasta': t.get('Cbte_vig_hasta')}
                tipos.append(tipo)
            except Exception as e:
                pass
        return ['%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in tipos]

    def GetParamNCM(self):
        "Recuperar lista de valores referenciales de c�digos del Nomenclador Com�n del Mercosur"
        ret = self.client.BFEGetPARAM_NCM(
            Auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetPARAM_NCMResult']
        self.__analizar_errores(result)

        ncms = [] # nomenclador comun del mercosur
        for n in result['BFEResultGet']:
            n = n['ClsBFEResponse_NCM']
            try:
                ncm = {'id': n.get('NCM_Codigo'), 'ds': n.get('NCM_Ds'),
                        'vig_desde': n.get('NCM_vig_desde'),
                        'vig_hasta': n.get('NCM_vig_hasta')}
                ncms.append(ncm)
            except Exception as e:
                pass
        return [u'%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in ncms]

    def GetParamZonas(self):
        "Recuperar lista de valores referenciales de Zonas"
        ret = self.client.BFEGetPARAM_Zonas(
            auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit, })
        result = ret['BFEGetPARAM_ZonasResult']
        self.__analizar_errores(result)

        zonas = [] # zonas
        for z in result['BFEResultGet']:
            z = z['ClsBFEResponse_Zon']
            try:
                zona = {'id': z.get('Zon_Id'), 'ds': z.get('Zon_Ds'),
                        'vig_desde': z.get('Zon_vig_desde'),
                        'vig_hasta': z.get('Zon_vig_hasta')}
                zonas.append(zona)
            except Exception as e:
                pass
        return ['%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in zonas]

    @inicializar_y_capturar_excepciones
    def GetParamCtz(self, moneda_id):
        "Recuperador de cotizaci�n de moneda"
        ret = self.client.BFEGetPARAM_Ctz(
            auth={'Token': self.Token, 'Sign': self.Sign, 'Cuit': self.Cuit},
            MonId=moneda_id,
            )
        self.__analizar_errores(ret['BFEGetPARAM_CtzResult'])
        res = ret['BFEGetPARAM_CtzResult'].get('BFEResultGet')
        if res:
            ctz = str(res.get('Mon_ctz',""))
        else:
            ctz = ''
        return ctz



class WSBFE(WSBFEv1):
    "Wrapper para retrocompatibilidad con WSBFE"

    _reg_progid_ = "WSBFE"
    _reg_clsid_ = "{02CBC6DA-455D-4EE6-8302-411D13253CBF}"

    def __init__(self):
        WSBFEv1.__init__(self)
        self.Version = "%s %s WSBFEv1" % (__version__, HOMO and 'Homologaci�n' or '')

    def Conectar(self, cache=None, url="", **kwargs):
        # Ajustar URL de V0 a V1:
        if url in ("https://wswhomo.afip.gov.ar/wsfex/service.asmx",
                   "http://wswhomo.afip.gov.ar/WSFEX/service.asmx"):
            url = "https://wswhomo.afip.gov.ar/wsbfev1/service.asmx"
        elif url in ("https://servicios1.afip.gov.ar/wsfex/service.asmx",
                     "http://servicios1.afip.gov.ar/WSFEX/service.asmx"):
            url = "https://servicios1.afip.gov.ar/wsbfev1/service.asmx"
        return WSBFEv1.Conectar(self, cache=cache, wsdl=url, **kwargs)


# busco el directorio de instalaci�n (global para que no cambie si usan otra dll)
INSTALL_DIR = WSBFEv1.InstallDir = get_install_dir()


def p_assert_eq(a,b):
    print(a, a==b and '==' or '!=', b)


if __name__ == "__main__":

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSBFEv1)
        if '--wsbfe' in sys.argv:
            win32com.server.register.UseCommandLine(WSBFE)
    else:

        # Crear objeto interface Web Service de Factura Electr�nica de Exportaci�n
        wsbfev1 = WSBFEv1()
        # Setear token y sing de autorizaci�n (pasos previos)

        # obteniendo el TA para pruebas
        from wsaa import WSAA
        ta = WSAA().Autenticar("wsbfe", "reingart.crt", "reingart.key")
        wsbfev1.SetTicketAcceso(ta)

        # CUIT del emisor (debe estar registrado en la AFIP)
        wsbfev1.Cuit = "20267565393"

        # Conectar al Servicio Web de Facturaci�n (homologaci�n)
        wsdl = "http://wswhomo.afip.gov.ar/WSBFEv1/service.asmx"
        cache = proxy = ""
        wrapper = "httplib2"
        # cacert = open("conf/afip_ca_info.crt").read()
        cacert = ""
        ok = wsbfev1.Conectar(cache, wsdl, proxy, wrapper, cacert)

        if '--dummy' in sys.argv:
            #wsbfev1.LanzarExcepciones = False
            print(wsbfev1.Dummy())
            print(wsbfev1.XmlRequest)
            print(wsbfev1.XmlResponse)

        if "--prueba" in sys.argv:
            try:
                # Establezco los valores de la factura a autorizar:
                tipo_cbte = '--nc' in sys.argv and 3 or 1 # FC/NC Expo (ver tabla de par�metros)
                punto_vta = 5
                tipo_doc = 80
                nro_doc = 23111111113
                zona = 0
                # Obtengo el �ltimo n�mero de comprobante y le agrego 1
                cbte_nro = int(wsbfev1.GetLastCMP(tipo_cbte, punto_vta)) + 1
                fecha_cbte = datetime.datetime.now().strftime("%Y%m%d")
                imp_moneda_id = "PES"  # (ver tabla de par�metros)
                imp_moneda_ctz = 1
                imp_neto = "390.00"
                impto_liq = "81.90"      # 21% IVA
                impto_liq_rni = imp_tot_conc = imp_op_ex = "0.00"
                imp_perc = imp_iibb = imp_perc_mun = imp_internos = "0.00"
                imp_total = "471.90"

                # Creo una factura (internamente, no se llama al WebService):
                ok = wsbfev1.CrearFactura(tipo_doc, nro_doc,
                    zona, tipo_cbte, punto_vta, cbte_nro, fecha_cbte,
                    imp_total, imp_neto, impto_liq,
                    imp_tot_conc, impto_liq_rni, imp_op_ex,
                    imp_perc, imp_iibb, imp_perc_mun, imp_internos,
                    imp_moneda_id, imp_moneda_ctz)

                # Agrego un item:
                ncm = '7308.10.00'
                sec = ''
                umed = 7  # unidades
                ds = 'prueba Anafe economico'
                qty = "2.00"
                precio = "100.00"
                bonif = "0.00"
                iva_id = 5
                imp_total = "242.00"
                ok = wsbfev1.AgregarItem(ncm, sec, ds, qty, umed, precio, bonif, iva_id, imp_total)

                # Agrego otro item:
                ncm = '7308.20.00'
                sec = ''
                umed = 7  # unidades
                ds = 'prueba 2'
                qty = "4.00"
                precio = "50.00"
                bonif = "10.00"
                iva_id = 5
                imp_total = "229.90"
                ok = wsbfev1.AgregarItem(ncm, sec, ds, qty, umed, precio, bonif, iva_id, imp_total)

                ##id = "99000000000100" # n�mero propio de transacci�n
                # obtengo el �ltimo ID y le adiciono 1
                # (advertencia: evitar overflow y almacenar!)
                id = long(wsbfev1.GetLastID()) + 1

                # Llamo al WebService de Autorizaci�n para obtener el CAE
                cae = wsbfev1.Authorize(id)

                print("Comprobante", tipo_cbte, wsbfev1.CbteNro)
                print("Resultado", wsbfev1.Resultado)
                print("CAE", wsbfev1.CAE)
                print("Vencimiento", wsbfev1.Vencimiento)

                if wsbfev1.Resultado and False:
                    print(wsbfev1.client.help("FEXGetCMP").encode("latin1"))
                    wsbfev1.GetCMP(tipo_cbte, punto_vta, cbte_nro)
                    print("CAE consulta", wsbfev1.CAE, wsbfev1.CAE == cae)
                    print("NRO consulta", wsbfev1.CbteNro,
                          wsbfev1.CbteNro == cbte_nro)
                    print("TOTAL consulta", wsbfev1.ImpTotal,
                          wsbfev1.ImpTotal == imp_total)

            except Exception as e:
                print(wsbfev1.XmlRequest)
                print(wsbfev1.XmlResponse)
                print(wsbfev1.ErrCode)
                print(wsbfev1.ErrMsg)
                print(wsbfev1.Excepcion)
                print(wsbfev1.Traceback)
                raise

        if "--get" in sys.argv:
            tipo_cbte = 1
            punto_vta = 5
            cbte_nro = wsbfev1.GetLastCMP(tipo_cbte, punto_vta)

            wsbfev1.GetCMP(tipo_cbte, punto_vta, cbte_nro)

            print("FechaCbte = ", wsbfev1.FechaCbte)
            print("CbteNro = ", wsbfev1.CbteNro)
            print("PuntoVenta = ", wsbfev1.PuntoVenta)
            print("ImpTotal =", wsbfev1.ImpTotal)
            print("CAE = ", wsbfev1.CAE)
            print("Vencimiento = ", wsbfev1.Vencimiento)

            wsbfev1.AnalizarXml("XmlResponse")
            p_assert_eq(wsbfev1.ObtenerTagXml('Cae'), str(wsbfev1.CAE))
            p_assert_eq(
                wsbfev1.ObtenerTagXml('Fecha_cbte_orig'), wsbfev1.FechaCbte
            )
            p_assert_eq(wsbfev1.ObtenerTagXml('Imp_moneda_Id'), "PES")
            p_assert_eq(wsbfev1.ObtenerTagXml('Imp_moneda_ctz'), "1")
            p_assert_eq(
                wsbfev1.ObtenerTagXml('Items', 'Item', 1, 'Pro_ds'), "prueba 2"
            )

        if "--params" in sys.argv:
            import codecs, locale
            sys.stdout = codecs.getwriter('latin1')(sys.stdout);

            print("=== Tipos de Comprobante ===")
            print(u'\n'.join(wsbfev1.GetParamTipoCbte()))

            print("=== Zonas ===")
            print(u'\n'.join(wsbfev1.GetParamZonas()))

            print("=== Monedas ===")
            print(u'\n'.join(wsbfev1.GetParamMon()))

            print("=== Tipos de IVA ===")
            print(u'\n'.join(wsbfev1.GetParamTipoIVA()))

            print("=== Unidades de medida ===")
            print(u'\n'.join(wsbfev1.GetParamUMed()))

            print(u"=== C�digos NCM ===")
            print(u'\n'.join(wsbfev1.GetParamNCM()))

        if "--ctz" in sys.argv:
            print(wsbfev1.GetParamCtz('DOL'))



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"M�dulo de Intefase para archivos de texto (exportaci�n)"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2010 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.24a"

import os
import sys
import time
# import traceback
from ConfigParser import SafeConfigParser

# revisar la instalaci�n de pyafip.ws:
import wsaa
import wsfex
from php import SimpleXMLElement, SoapClient, SoapFault, date


HOMO = False
DEBUG = False
XML = False
CONFIG_FILE = "rece.ini"

LICENCIA = """
recex.py: Interfaz de texto para generar Facturas Electr�nica Exportaci�n
Copyright (C) 2010 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para informaci�n adicional sobre garant�a, soporte t�cnico comercial
e incorporaci�n/distribuci�n en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

# definici�n del formato del archivo de intercambio:
N = 'Numerico'
A = 'Alfanumerico'
I = 'Importe'
ENCABEZADO = [
    ('tipo_reg', 1, N),  # 0: encabezado
    ('fecha_cbte', 8, A),
    ('tipo_cbte', 2, N), ('punto_vta', 4, N),
    ('cbte_nro', 8, N),
    ('tipo_expo', 1, N),  # 1:bienes, 2:servicios,...
    ('permiso_existente', 1, A),  # S/N/
    ('dst_cmp', 3, N),  # 203
    ('cliente', 200, A),  # 'Joao Da Silva'
    ('cuit_pais_cliente', 11, N),  # 50000000016
    ('domicilio_cliente', 300, A),  # 'Rua 76 km 34.5 Alagoas'
    ('id_impositivo', 50, A),  # 'PJ54482221-l'
    ('imp_total', 15, I, 3),
    ('moneda_id', 3, A),
    ('moneda_ctz', 10, I, 6),  # 10,6
    ('obs_comerciales', 1000, A),
    ('obs', 1000, A),
    ('forma_pago', 50, A),
    ('incoterms', 3, A),
    ('incoterms_ds', 20, A),
    ('idioma_cbte', 1, A),
    ('cae', 14, N), ('fecha_vto', 8, A),
    ('resultado', 1, A),
    ('reproceso', 1, A),
    ('motivos_obs', 40, A),
    ('id', 15, N),
    ('fch_venc_cae', 8, A),
    ]

DETALLE = [
    ('tipo_reg', 1, N),  # 1: detalle item
    ('codigo', 30, A),
    ('qty', 12, I),
    ('umed', 2, N),
    ('precio', 12, I, 3),
    ('imp_total', 14, I, 3),
    ('ds', 4000, A),
    ]

PERMISO = [
    ('tipo_reg', 1, N),  # 2: permiso
    ('id_permiso', 16, A),
    ('dst_merc', 3, N),
    ]

CMP_ASOC = [
    ('tipo_reg', 1, N),  # 3: comprobante asociado
    ('cbte_tipo', 3, N), ('cbte_punto_vta', 4, N),
    ('cbte_nro', 8, N),
    ]


def leer(linea, formato):
    dic = {}
    comienzo = 1
    for fmt in formato:
        clave, longitud, tipo = fmt[0:3]
        dec = len(fmt) > 3 and fmt[3] or 2
        valor = linea[comienzo-1:comienzo-1+longitud].strip()
        try:
            if tipo == N:
                if valor:
                    valor = str(int(valor))
                else:
                    valor = '0'
            elif tipo == I:
                if valor:
                    try:
                        valor = valor.strip(" ")
                        valor = float(
                            ("%%s.%%0%sd" % dec) % (
                                int(valor[:-dec] or '0'),
                                int(valor[-dec:] or '0'))
                        )
                    except ValueError:
                        raise ValueError(
                            "Campo invalido: %s = '%s'" % (clave, valor))
                else:
                    valor = 0.00
            else:
                valor = valor.decode("ascii", "ignore")
            dic[clave] = valor

            comienzo += longitud
        except Exception as e:
            raise ValueError("Error al leer campo %s pos %s val '%s': %s" % (
                clave, comienzo, valor, str(e)))
    return dic


translate_keys = {
    'moneda_id': 'Moneda_Id',
    'codigo': 'Pro_codigo', 'precio': 'Pro_precio_uni',
    'imp_total': 'Pro_total_item', 'ds': 'Pro_ds', 'umed': 'Pro_umed',
    'qty': 'Pro_qty', }


def escribir(dic, formato):
    linea = " " * 335
    comienzo = 1
    for fmt in formato:
        clave, longitud, tipo = fmt[0:3]
        dec = len(fmt) > 3 and fmt[3] or 2
        if clave.capitalize() in dic:
            clave = clave.capitalize()
        valor = str(dic.get(clave, ""))
        if valor == "" and clave in translate_keys:
            valor = str(dic.get(translate_keys[clave], ""))
        if tipo == N and valor and valor != "NULL":
            valor = ("%%0%dd" % longitud) % int(valor)
        elif tipo == I and valor:
            valor = ("%%0%dd" % longitud) % (float(valor)*(10**dec))
        else:
            valor = ("%%0%ds" % longitud) % valor
        linea = linea[:comienzo-1] + valor + linea[comienzo-1+longitud:]
        comienzo += longitud
    return linea + "\n"


def autenticar(cert, privatekey, url):
    "Obtener el TA"
    TA = "TA.xml"
    ttl = 60*60*5
    if not os.path.exists(TA) or os.path.getmtime(TA)+(ttl) < time.time():
        import wsaa
        tra = wsaa.create_tra(service="wsfex", ttl=ttl)
        cms = wsaa.sign_tra(str(tra), str(cert), str(privatekey))
        ta_string = wsaa.call_wsaa(cms, wsaa_url, trace=DEBUG)
        open(TA, "w").write(ta_string)
    ta_string = open(TA).read()
    ta = SimpleXMLElement(ta_string)
    token = str(ta.credentials.token)
    sign = str(ta.credentials.sign)
    return token, sign


def autorizar(client, token, sign, cuit, entrada, salida):
    # recupero el �ltimo n�mero de transacci�n
    # id = wsfex.ultnro(client, token, sign, cuit)

    detalles = []
    permisos = []
    cbtasocs = []
    encabezado = {}
    for linea in entrada:
        if str(linea[0]) == '0':
            encabezado = leer(linea, ENCABEZADO)
        elif str(linea[0]) == '1':
            detalle = leer(linea, DETALLE)
            detalles.append(detalle)
        elif str(linea[0]) == '2':
            permiso = leer(linea, PERMISO)
            permisos.append(permiso)
        elif str(linea[0]) == '3':
            cbtasoc = leer(linea, CMP_ASOC)
            cbtasocs.append(cbtasoc)
        else:
            print("Tipo de registro incorrecto:", linea[0])

    if not encabezado['id'].strip() or int(encabezado['id']) == 0:
        # TODO: habria que leer y/o grabar el id en el archivo
        # id += 1 # incremento el n� de transacci�n
        # Por el momento, el id se calcula con el tipo, pv y n� de comprobant
        i = long(encabezado['cbte_nro'])
        i += (
            int(encabezado['cbte_nro'])*10**4 + int(
                encabezado['punto_vta'])
        )*10**8
        encabezado['id'] = wsfex.get_last_id(client, token, sign, cuit)[0] + 1

    factura = wsfex.FacturaEX(**encabezado)
    for detalle in detalles:
        it = wsfex.ItemFEX(**detalle)
        factura.add_item(it, calcular=False)
    for permiso in permisos:
        p = wsfex.PermisoFEX(**permiso)
        factura.add_permiso(p)
    for cbtasoc in cbtasocs:
        c = wsfex.CmpAsocFEX(**cbtasoc)
        factura.add_cmp_asoc(c)

    if DEBUG:
        print(f.to_dict())
        print(
            '\n'.join(
                ["%s='%s'" % (k, str(v)) for k, v in factura.to_dict().items()]
            )
        )
        print('id:', encabezado['id'])

    if not DEBUG or raw_input("Facturar?") == "S":
        dic = factura.to_dict()
        auth, events = wsfex.authorize(client, token, sign, cuit,
                                       id=encabezado['id'],
                                       factura=dic)
        dic.update(auth)
        escribir_factura(dic, salida)
        print(
            "ID:", dic['id'], "CAE:", dic['cae'], "Obs:", dic['obs'],
            "Reproceso:", dic['reproceso'])


def escribir_factura(dic, archivo):
    dic['tipo_reg'] = 0
    archivo.write(escribir(dic, ENCABEZADO))
    for it in dic['Items']:
        it['Item']['tipo_reg'] = 1
        archivo.write(escribir(it['Item'], DETALLE))
    if 'Permisos' in dic:
        for it in dic['Permisos']:
            it['Permiso']['tipo_reg'] = 2
            archivo.write(escribir(it['Permiso'], PERMISO))


def depurar_xml(client):
    fecha = time.strftime("%Y%m%d%H%M%S")
    f = open("request-%s.xml" % fecha, "w")
    f.write(client.xml_request)
    f.close()
    f = open("response-%s.xml" % fecha, "w")
    f.write(client.xml_response)
    f.close()


if __name__ == "__main__":
    if '/ayuda' in sys.argv:
        print(LICENCIA)
        print("Opciones: ")
        print(" /ayuda: este mensaje")
        print(" /dummy: consulta estado de servidores")
        print(
            " /prueba: genera y autoriza una factura de prueba \
            (no usar en producci�n!)")
        print(" /ult: consulta �ltimo n�mero de comprobante")
        print(" /debug: modo depuraci�n (detalla y confirma las operaciones)")
        print(
            " /formato: muestra el formato de los archivos de entrada/salida")
        print(
            " /get: recupera datos de un comprobante autorizado previamente \
            (verificaci�n)")
        print(
            " /xml: almacena los requerimientos y respuestas XML \
            (depuraci�n)")
        print(
            "Ver rece.ini para par�metros de configuraci�n \
            (URL, certificados, etc.)")
        sys.exit(0)

    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    cert = config.get('WSAA', 'CERT')
    privatekey = config.get('WSAA', 'PRIVATEKEY')
    cuit = config.get('WSFEX', 'CUIT')
    entrada = config.get('WSFEX', 'ENTRADA')
    salida = config.get('WSFEX', 'SALIDA')

    if config.has_option('WSAA', 'URL') and not HOMO:
        wsaa_url = config.get('WSAA', 'URL')
    else:
        wsaa_url = wsaa.WSAAURL
    if config.has_option('WSFEX', 'URL') and not HOMO:
        wsfex_url = config.get('WSFEX', 'URL')
    else:
        wsfex_url = wsfex.WSFEXURL

    if '/debug'in sys.argv:
        DEBUG = True

    if '/xml'in sys.argv:
        XML = True

    if DEBUG:
        print("wsaa_url %s\nwsfex_url %s" % (wsaa_url, wsfex_url))

    try:
        client = SoapClient(
            wsfex_url, action=wsfex.SOAP_ACTION, namespace=wsfex.SOAP_NS,
            trace=DEBUG, exceptions=True)

        if '/dummy' in sys.argv:
            print("Consultando estado de servidores...")
            print(wsfex.dummy(client))
            sys.exit(0)

        if '/formato' in sys.argv:
            print("Formato:")
            for msg, formato in [
                ('Encabezado', ENCABEZADO),
                ('Detalle', DETALLE),
                ('Permiso', PERMISO),
                ('Comprobante Asociado', CMP_ASOC)
            ]:
                comienzo = 1
                print("== %s ==" % msg)
                for fmt in formato:
                    clave, longitud, tipo = fmt[0:3]
                    dec = len(fmt) > 3 and fmt[3] or \
                        (tipo == 'I' and '2' or '')
                    print(
                        " * Campo: %-20s Posici�n: %3d Longitud: %4d Tipo: \
                        %s Decimales: %s" % (
                            clave, comienzo, longitud, tipo, dec)
                    )
                    comienzo += longitud
            sys.exit(0)

        # TODO: esto habr�a que guardarlo en un archivo y no tener que
        # autenticar cada vez
        token, sign = autenticar(cert, privatekey, wsaa_url)

        if '/prueba' in sys.argv:
            # generar el archivo de prueba para la pr�xima factura
            fecha = date('Ymd')
            tipo_cbte = 19
            punto_vta = 3
            ult_cbte, fecha, events = wsfex.get_last_cmp(
                client, token, sign, cuit, tipo_cbte, punto_vta)

            ult_id, events = wsfex.get_last_id(client, token, sign, cuit)

            f_entrada = open(entrada, "w")

            f = wsfex.FacturaEX()
            f.punto_vta = punto_vta
            f.cbte_nro = ult_cbte+1
            f.imp_moneda_id = 'PES'
            f.fecha_cbte = date('Ymd')
            f.tipo_expo = 1
            f.permiso_existente = 'S'
            f.dst_cmp = 203
            f.cliente = 'Joao Da Silva'
            f.cuit_pais_cliente = 50000000016
            f.domicilio_cliente = 'Rua 76 km 34.5 Alagoas'
            f.id_impositivo = 'PJ54482221-l'
            f.moneda_id = 'DOL'
            f.moneda_ctz = 3.875
            f.obs_comerciales = 'Observaciones comerciales'
            f.obs = 'Sin observaciones'
            f.forma_pago = '30 dias'
            f.incoterms = 'FOB'
            f.idioma_cbte = 1
            # agrego los items
            it = wsfex.ItemFEX(
                codigo='PRO1',
                ds=u'Producto Tipo 1 Exportacion MERCOSUR ISO 9001', qty=1,
                precio=125)
            f.add_item(it)
            it = wsfex.ItemFEX(
                codigo='PRO2',
                ds=u'Producto Tipo 2 Exportacion MERCOSUR ISO 9001', qty=1,
                precio=125)
            f.add_item(it)
            permiso = wsfex.PermisoFEX(
                id_permiso="99999AAXX999999A", dst_merc=225)
            f.add_permiso(permiso)

            if DEBUG:
                print(f.to_dict())

            dic = f.to_dict()
            dic['id'] = ult_id+1
            escribir_factura(dic, f_entrada)
            f_entrada.close()

        if '/ult' in sys.argv:
            print("Consultar ultimo numero:")
            tipo_cbte = int(raw_input("Tipo de comprobante: "))
            punto_vta = int(raw_input("Punto de venta: "))
            ult_cbte, fecha, events = wsfex.get_last_cmp(
                client, token, sign, cuit, tipo_cbte, punto_vta)
            print("Ultimo numero: ", ult_cbte)
            print("Fecha: ", fecha)
            depurar_xml(client)
            sys.exit(0)

        if '/get' in sys.argv:
            print("Recuperar comprobante:")
            tipo_cbte = int(raw_input("Tipo de comprobante: "))
            punto_vta = int(raw_input("Punto de venta: "))
            cbte_nro = int(raw_input("Numero de comprobante: "))
            cbt, events = wsfex.get_cmp(
                client, token, sign, cuit, tipo_cbte, punto_vta, cbte_nro)
            for k, v in cbt.items():
                print("%s = %s" % (k, v))
            depurar_xml(client)
            sys.exit(0)

        f_entrada = f_salida = None
        try:
            f_entrada = open(entrada, "r")
            f_salida = open(salida, "w")
            try:
                autorizar(client, token, sign, cuit, f_entrada, f_salida)
            except (SoapFault, wsfex.FEXError) as e:
                XML = True
                raise
        finally:
            if f_entrada is not None:
                f_entrada.close()
            if f_salida is not None:
                f_salida.close()
            if XML:
                depurar_xml(client)
        sys.exit(0)

    except SoapFault as e:
        print(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        sys.exit(3)
    except wsfex.FEXError as e:
        print(e.code, e.msg.encode("ascii", "ignore"))
        sys.exit(4)
    except Exception as e:
        print(unicode(e).encode("ascii", "ignore"))
        if DEBUG:
            raise
        sys.exit(5)
